"""Tests for the CDS GeoBoundingBox module."""
import pytest
from geopandas import GeoSeries
from shapely.geometry import Polygon

from aatoolbox.utils.geoboundingbox import GeoBoundingBox


def test_geoboundingbox_round_coords():
    """Test that coordinates are correctly rounded."""
    geobb = GeoBoundingBox(
        lat_max=1.5, lat_min=-2.2, lon_max=3.6, lon_min=-4.0
    ).round_coords(round_val=0.5)
    assert geobb.lat_max == 1.5
    assert geobb.lat_min == -2.5
    assert geobb.lon_max == 4.0
    assert geobb.lon_min == -4.0


def test_geoboundingbox_offset_coords():
    """Test that coordinates are correctly offset."""
    geobb = GeoBoundingBox(
        lat_max=1.05, lat_min=-2.2, lon_max=3.6, lon_min=-4.0
    ).round_coords(
        offset_val=0.4
    )  # default round val of 1
    assert geobb.lat_max == 2.4
    assert geobb.lat_min == -3.4
    assert geobb.lon_max == 4.4
    assert geobb.lon_min == -4.4


def test_geoboundingbox_round_and_offset_coords():
    """Test that coordinates are correctly rounded and offset."""
    geobb = GeoBoundingBox(
        lat_max=1.05, lat_min=-2.2, lon_max=3.6, lon_min=-4.0
    ).round_coords(round_val=0.1, offset_val=0.05)
    assert geobb.lat_max == 1.15
    assert geobb.lat_min == -2.25
    assert geobb.lon_max == 3.65
    assert geobb.lon_min == -4.05


def test_geoboundingbox_relative_coords():
    """Test that lat_max must be >= lat_min and lon_max >= lon_min."""
    with pytest.raises(AttributeError):
        # Check lat_min > lat_max
        GeoBoundingBox(lat_max=1, lat_min=2, lon_max=1, lon_min=-1)
        # Check lon_min > lon_max
        GeoBoundingBox(lat_max=1, lat_min=-1, lon_max=1, lon_min=2)
    # Check that no error raised:
    # Check lat_min == lat_max
    GeoBoundingBox(lat_max=1, lat_min=1, lon_max=1, lon_min=-1)
    # Check lon_min == lon_max
    GeoBoundingBox(lat_max=1, lat_min=-1, lon_max=1, lon_min=1)


def test_geoboundingbox_coordinate_bounds():
    """Test that coords are within limits."""
    with pytest.raises(AttributeError):
        GeoBoundingBox(lat_max=91, lat_min=0, lon_max=0, lon_min=0)
        GeoBoundingBox(lat_max=0, lat_min=-91, lon_max=0, lon_min=0)
        GeoBoundingBox(lat_max=0, lat_min=0, lon_max=181, lon_min=0)
        GeoBoundingBox(lat_max=0, lat_min=0, lon_max=0, lon_min=-181)


def test_geoboundingbox_filename():
    """Test that correct file string is returned."""
    geobb = GeoBoundingBox(lat_max=1, lat_min=-2, lon_max=3, lon_min=-4)
    assert geobb.get_filename_repr() == "Np1Sm2Ep3Wm4"


def test_geoboundingbox_from_shape():
    """Test that getting the geobb from a polygon works as expected."""
    n, s, e, w = (1, -2, 3, -4)
    shape = GeoSeries([Polygon([(e, n), (e, s), (w, s), (w, n)])])
    geobb = GeoBoundingBox.from_shape(shape)
    assert geobb.lat_max == n
    assert geobb.lat_min == s
    assert geobb.lon_max == e
    assert geobb.lon_min == w



"""Tests for the USGS NDVI module."""
import tempfile
from datetime import date, datetime, timedelta
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from shapely.geometry import Polygon

from aatoolbox import (
    UsgsNdviMedianAnomaly,
    UsgsNdviPctMedian,
    UsgsNdviSmoothed,
    UsgsNdviYearDifference,
)

DATASOURCE_BASE_DIR = "usgs_ndvi"


@pytest.fixture
def mock_ndvi(mock_country_config):
    """Create USGS NDVI class with mock country config."""
    start_date = (2019, 36)
    end_date = (2020, 2)

    instantiator = {
        "smoothed": UsgsNdviSmoothed,
        "pct_median": UsgsNdviPctMedian,
        "anomaly": UsgsNdviMedianAnomaly,
        "difference": UsgsNdviYearDifference,
    }

    def _mock_ndvi(variable: str = "smoothed"):

        ndvi = instantiator[variable](
            country_config=mock_country_config,
            start_date=start_date,
            end_date=end_date,
        )
        return ndvi

    return _mock_ndvi


@pytest.fixture
def mock_download(mocker, mock_ndvi):
    """
    Call download with mocked download.

    Return number of calls to the internal
    download method and the returned filepath.
    """
    download_mock = mocker.patch(
        (
            "aatoolbox.datasources.usgs."
            "ndvi_base._UsgsNdvi._download_ndvi_dekad"
        )
    )

    def _mock_download(variable: str = "smoothed"):
        ndvi = mock_ndvi(variable=variable)
        fp = ndvi.download()
        download_calls = download_mock.call_count
        return fp, download_calls

    return _mock_download


@pytest.fixture
def mock_determine_process_dates(mocker, mock_ndvi):
    """
    Call _determine_process_dates() with mocked _load().

    Return number of calls to the internal
    download method and the returned filepath.
    """
    mocker.patch(
        ("aatoolbox.datasources.usgs.ndvi_base._UsgsNdvi._load"),
        return_value=pd.DataFrame(
            {
                "year": [2019, 2020, 2020],
                "dekad": [36, 1, 2],
                "modified": [datetime.now() + timedelta(days=10)] * 3,
            }
        ),
    )

    mocker.patch(
        "aatoolbox.datasources.usgs.ndvi_base._UsgsNdvi._get_raw_path",
        return_value=Path(tempfile.mkstemp()[1]),
    )

    def _mock_determine_process_dates(clobber: bool, dates_to_process: list):
        ndvi = mock_ndvi(variable="smoothed")
        return ndvi._determine_process_dates(
            clobber=clobber,
            filepath="fake_filepath",
            dates_to_process=dates_to_process,
        )

    return _mock_determine_process_dates


@pytest.fixture
def gdf():
    """Create geodataframe for testing."""
    d = {
        "name": ["area_a", "area_b"],
        "geometry": [
            Polygon([(0, 0), (0, 2), (2, 2), (2, 0)]),
            Polygon([(2, 0), (2, 2), (3, 2), (3, 0)]),
        ],
    }
    return gpd.GeoDataFrame(d)


@pytest.fixture
def da():
    """
    Create raster input with date coordinate.

    Used as side effect later to create 3
    consecutive raster values for processing.
    """

    def _da(year_dekad):
        da = xr.DataArray(
            [[[1 + year_dekad[1], 2, 3], [4, 5 + year_dekad[1], 6]]],
            dims=("band", "y", "x"),
            coords={
                "y": [1.5, 0.5],
                "x": [0.5, 1.5, 2.5],
            },
        ).rio.write_crs("EPSG:4326")
        return da

    return _da


def test_download_smoothed(mock_download, mock_aa_data_dir):
    """Test download for all NDVI variables."""
    fp, call_count = mock_download(variable="smoothed")

    assert fp == (mock_aa_data_dir / f"public/raw/glb/{DATASOURCE_BASE_DIR}")
    assert call_count == 3


def test_download_pct_median(mock_download, mock_aa_data_dir):
    """Test download for all NDVI variables."""
    fp, call_count = mock_download(variable="pct_median")

    assert fp == (mock_aa_data_dir / f"public/raw/glb/{DATASOURCE_BASE_DIR}")

    assert call_count


def test_download_anomaly(mock_download, mock_aa_data_dir):
    """Test download for all NDVI variables."""
    fp, call_count = mock_download(variable="anomaly")

    assert fp == (mock_aa_data_dir / f"public/raw/glb/{DATASOURCE_BASE_DIR}")

    assert call_count


def test_download_difference(mock_download, mock_aa_data_dir):
    """Test download for all NDVI variables."""
    fp, call_count = mock_download(variable="difference")

    assert fp == (mock_aa_data_dir / f"public/raw/glb/{DATASOURCE_BASE_DIR}")

    assert call_count


def test_process_and_load(
    mocker, mock_ndvi, mock_aa_data_dir, mock_country_config, gdf, da
):
    """Test processing NDVI values."""
    ndvi = mock_ndvi(variable="smoothed")

    # TODO: now created `load_raw` to be able to mock but would like
    #  to do it from xr.load_dataset directly
    mocker.patch(
        "aatoolbox.datasources.usgs.ndvi_base.rioxarray.open_rasterio",
        side_effect=[da((2019, 36)), da((2020, 1)), da((2020, 2))] * 6,
    )

    mocker.patch(
        "aatoolbox.datasources.usgs.ndvi_base._UsgsNdvi._get_raw_path",
        return_value=Path(tempfile.mkstemp()[1]),
    )

    processed_dir = ndvi.process(gdf=gdf, feature_col="name")
    assert processed_dir == (
        mock_aa_data_dir / f"public/processed/{mock_country_config.iso3}/"
        f"{DATASOURCE_BASE_DIR}"
    )

    df_processed = ndvi.load(feature_col="name")

    expected_dates = [
        date(year=2019, month=12, day=21),
        date(year=2020, month=1, day=1),
        date(year=2020, month=1, day=11),
    ]

    expected_dates = [d for ds in expected_dates for d in [ds] * 2]

    assert np.array_equal(df_processed.name.values, ["area_a", "area_b"] * 3)
    assert np.array_equal(
        pd.to_datetime(df_processed.date).dt.date, expected_dates
    )
    assert np.array_equal(df_processed.dekad.values, [36, 36, 1, 1, 2, 2])
    assert np.array_equal(df_processed.loc[:, "count"].values, [4, 2] * 3)
    assert np.array_equal(
        df_processed.loc[:, "min"].values, [2, 3, 2, 3, 2, 3]
    )
    assert np.allclose(
        df_processed.loc[:, "std"].values,
        [18.069311, 1.5, 1.6583124, 1.5, 1.87082869, 1.5],
    )
    assert np.array_equal(
        df_processed.loc[:, "mean"].values, [21.0, 4.5, 3.5, 4.5, 4.0, 4.5]
    )
    assert np.array_equal(
        df_processed.loc[:, "max"].values, [41, 6, 6, 6, 7, 6]
    )


def test_process_if_download_not_called(mock_ndvi, gdf):
    """Test that correct error message raised."""
    ndvi = mock_ndvi(variable="smoothed")
    with pytest.raises(FileNotFoundError):
        ndvi.process(gdf=gdf, feature_col="name")


def test_load_if_process_not_called(mock_ndvi):
    """Test that correct error message raised."""
    ndvi = mock_ndvi(variable="smoothed")
    with pytest.raises(FileNotFoundError):
        ndvi.load(feature_col="name")


def test_fp_year_dekad():
    """Test that year and dekad extracted from FP."""
    fp = Path("ea2022_10pct.tif")
    assert UsgsNdviPctMedian._fp_year_dekad(fp) == [2022, 10]


def test_get_url(mock_ndvi):
    """Test that URL generated correctly."""
    fn = "fake_filename"
    ndvi = mock_ndvi(variable="pct_median")
    expected_fn = (
        f"https://edcintl.cr.usgs.gov/downloads/sciweb1/shared/fews/web/"
        f"{ndvi._datasource_config.area_url}/dekadal/emodis"
        f"/ndvi_c6/{ndvi._data_variable_url}/"
        f"downloads/dekadal/{fn}.zip"
    )
    assert ndvi._get_url(fn) == expected_fn


def test_process_dates_clobber_true(mock_determine_process_dates):
    """Test process dates clobbers properly."""
    dps = [(2019, 36), (2020, 1), (2020, 2)]
    test_dps, df = mock_determine_process_dates(
        clobber=True, dates_to_process=dps
    )
    assert len(test_dps) == 3
    assert df.size == 0


def test_process_dates_clobber_false(mock_determine_process_dates):
    """Test process dates doesn't clobber unnecessarily."""
    dps = [(2019, 36), (2020, 1), (2020, 2)]
    test_dps, df = mock_determine_process_dates(
        clobber=False, dates_to_process=dps
    )
    assert len(test_dps) == 0
    assert df.shape[0] == 3


def test_process_dates_clobber_true_additional(mock_determine_process_dates):
    """Test process dates clobbers properly."""
    dps = [(2019, 35), (2019, 36), (2020, 1), (2020, 2), (2020, 3)]
    test_dps, df = mock_determine_process_dates(
        clobber=True, dates_to_process=dps
    )
    assert len(test_dps) == 5
    assert df.size == 0


def test_process_dates_clobber_false_additional(mock_determine_process_dates):
    """Test process dates doesn't clobber unnecessarily."""
    dps = [(2019, 35), (2019, 36), (2020, 1), (2020, 2), (2020, 3)]
    test_dps, df = mock_determine_process_dates(
        clobber=False, dates_to_process=dps
    )
    assert len(test_dps) == 2
    assert df.shape[0] == 3



"""aa-toolbox: Toolbox for Anticipatory Action.

The ``aatoolbox`` library contains modules designed to assist with
downloading and processing the data required for disaster-related
anticipatory action.
"""
from aatoolbox.config.countryconfig import (
    CountryConfig,
    create_country_config,
    create_custom_country_config,
)
from aatoolbox.datasources.chirps.chirps import ChirpsDaily, ChirpsMonthly
from aatoolbox.datasources.codab.codab import CodAB
from aatoolbox.datasources.fewsnet.fewsnet import FewsNet
from aatoolbox.datasources.glofas.forecast import (
    GlofasForecast,
    GlofasReforecast,
)
from aatoolbox.datasources.glofas.reanalysis import GlofasReanalysis
from aatoolbox.datasources.iri.iri_seasonal_forecast import (
    IriForecastDominant,
    IriForecastProb,
)
from aatoolbox.datasources.usgs.ndvi_products import (
    UsgsNdviMedianAnomaly,
    UsgsNdviPctMedian,
    UsgsNdviSmoothed,
    UsgsNdviYearDifference,
)
from aatoolbox.utils import raster
from aatoolbox.utils.geoboundingbox import GeoBoundingBox

from ._version import version as __version__

__all__ = (
    "CountryConfig",
    "create_country_config",
    "create_custom_country_config",
    "ChirpsDaily",
    "ChirpsMonthly",
    "CodAB",
    "FewsNet",
    "GlofasReanalysis",
    "GlofasForecast",
    "GlofasReforecast",
    "IriForecastProb",
    "IriForecastDominant",
    "UsgsNdviSmoothed",
    "UsgsNdviPctMedian",
    "UsgsNdviMedianAnomaly",
    "UsgsNdviYearDifference",
    "GeoBoundingBox",
    "raster",
    "__version__",
)



"""Functions for dealing with dates."""

import itertools
from datetime import date
from typing import List, Tuple, Union, cast


def get_date_from_user_input(input_date: Union[date, str]) -> date:
    """Return date from string or date input.

    Processes input data in either ``datetime.date``
    format or as an ISO8601 string. Generates
    error message if different object provided.

    Parameters
    ----------
    input_date : Union[date, str]
        ``datetime.date`` object or ISO8601 string.

    Returns
    -------
    date
        ``datetime.date``
    """
    if isinstance(input_date, date):
        return input_date
    try:
        input_date = date.fromisoformat(input_date)
    except ValueError as err:
        raise ValueError(
            "`date` values passed as a string must follow ISO8601 date "
            "format: YYYY-MM-DD."
        ) from err
    except TypeError as err:
        raise TypeError(
            "`date` values must be either an ISO8601 string or "
            "`datetime.date` object."
        ) from err

    return input_date


def get_dekadal_date(
    input_date: Union[date, str, Tuple[int, int], None],
    default_date: Union[date, str, Tuple[int, int], None] = None,
) -> Tuple[int, int]:
    """Calculate dekadal date from general input.

    Processes input ``input_date`` and returns two
    values, the year and dekad. Input can be of
    format ``datetime.date``, an ISO8601 date
    string, an already calculated ``(year, dekad)``
    format date, or ``None``. If ``None``,
    ``default_date`` is returned. ``default_date``
    can also be passed in the above formats.
    """
    if input_date is None and default_date is not None:
        input_date = default_date

    # convert date to various values
    if not isinstance(input_date, (str, date)):
        input_tuple = cast(Tuple[int, int], input_date)
        if len(input_tuple) == 2:
            year, dekad = input_tuple
            # assert year-dekad values appropriate, not too strict
            if year < 1000 or year > 9999 or dekad < 1 or dekad > 36:
                raise ValueError(
                    f"(year, dekad) tuple ({year}, {dekad}) invalid. "
                    "Year should be a 4-digit year and dekad between "
                    "1 and 36."
                )

        else:
            raise ValueError(
                (
                    "`date` values for dekadal data "
                    "should be passed in as "
                    "`datetime.date` objects, tuples "
                    "of `(year, dekad)` format, or "
                    "ISO8601 date strings."
                )
            )

    else:
        input_as_date = get_date_from_user_input(input_date)
        year, dekad = date_to_dekad(input_as_date)

    return year, dekad


def dekad_to_date(dekad: Tuple[int, int]) -> date:
    """Compute date from dekad and year.

    Date computed from dekad and year in
    datetime object, corresponding to
    first day of the dekad. This
    is based on the
    `common dekadal definition
    <http://iridl.ldeo.columbia.edu/maproom/Food_Security/Locusts/Regional/Dekadal_Rainfall/index.html>`_
    of the 1st and 2nd dekad of a month
    being the first 10 day periods, and
    the 3rd dekad being the remaining
    days within that month.
    """
    year = dekad[0]
    month = ((dekad[1] - 1) // 3) + 1
    day = 10 * ((dekad[1] - 1) % 3) + 1
    return date(year=year, month=month, day=day)


def date_to_dekad(date_obj: date) -> Tuple[int, int]:
    """Compute dekad and year from date.

    Dekad computed from date. This
    is based on the
    `common dekadal definition
    <http://iridl.ldeo.columbia.edu/maproom/Food_Security/Locusts/Regional/Dekadal_Rainfall/index.html>`_
    of the 1st and 2nd dekad of a month
    being the first 10 day periods, and
    the 3rd dekad being the remaining
    days within that month.
    """
    year = date_obj.year
    dekad = min((date_obj.day - 1) // 10, 2) + ((date_obj.month - 1) * 3) + 1
    return (year, dekad)


def compare_dekads_lt(
    dekad1: Tuple[int, int], dekad2: Tuple[int, int]
) -> bool:
    """Is year1/dekad1 less than year2/dekad2.

    Compare two pairs of years and dekads,
    that the first pair are less than the
    second pair.
    """
    y1, d1 = dekad1
    y2, d2 = dekad2
    return y1 < y2 or ((y1 == y2) and (d1 < d2))


def compare_dekads_lte(
    dekad1: Tuple[int, int], dekad2: Tuple[int, int]
) -> bool:
    """Is year1/dekad1 less than or equal to year2/dekad2.

    Compare two pairs of years and dekads,
    that the first pair are less than or
    equal to the second pair.
    """
    y1, d1 = dekad1
    y2, d2 = dekad2
    return y1 < y2 or ((y1 == y2) and (d1 <= d2))


def compare_dekads_gt(
    dekad1: Tuple[int, int], dekad2: Tuple[int, int]
) -> bool:
    """Is year1/dekad1 greater than year2/dekad2.

    Compare two pairs of years and dekads,
    that the first pair are greater than the
    second pair.
    """
    return compare_dekads_lt(dekad1=dekad2, dekad2=dekad1)


def compare_dekads_gte(
    dekad1: Tuple[int, int], dekad2: Tuple[int, int]
) -> bool:
    """Is year1/dekad1 greater than or equal to year2/dekad2.

    Compare two pairs of years and dekads,
    that the first pair are greater than or
    equal to the second pair.
    """
    return compare_dekads_lte(dekad1=dekad2, dekad2=dekad1)


def expand_dekads(
    dekad1: Tuple[int, int], dekad2: Tuple[int, int]
) -> List[Tuple[int, int]]:
    """Expand for all years/dekads between two dates.

    Takes input year and dekads and returns a list
    of year/dekad lists.
    """
    if compare_dekads_gt(dekad1, dekad2):
        raise ValueError("`dekad1` must be less than or equal to `dekad2`.")

    y1, d1 = dekad1
    y2, d2 = dekad2
    year_range = range(y1, y2 + 1)
    dekad_range = range(1, 37)
    date_combos = itertools.product(*[year_range, dekad_range])

    def valid(y, d):
        return not ((y == y1 and d < d1) or (y == y2 and d > d2))

    return [(y, d) for y, d in date_combos if valid(y, d)]



"""Download and process GloFAS forecast and reforecast river discharge data."""
import logging
from datetime import date
from pathlib import Path
from typing import List, Union

import xarray as xr
from dateutil import rrule

from aatoolbox.config.countryconfig import CountryConfig
from aatoolbox.datasources.glofas import glofas
from aatoolbox.utils.check_file_existence import check_file_existence
from aatoolbox.utils.geoboundingbox import GeoBoundingBox

logger = logging.getLogger(__name__)


class _GlofasForecastBase(glofas.Glofas):
    """Base class for all GloFAS forecast data downloading and processing."""

    def __init__(
        self,
        country_config: CountryConfig,
        geo_bounding_box: GeoBoundingBox,
        cds_name: str,
        system_version: str,
        product_type: Union[str, List[str]],
        date_variable_prefix: str,
        frequency: int,
        coord_names: List[str],
        leadtime_max: int,
        start_date_min: date,
        end_date_max: date = None,
        start_date: Union[date, str] = None,
        end_date: Union[date, str] = None,
    ):
        super().__init__(
            country_config=country_config,
            geo_bounding_box=geo_bounding_box,
            cds_name=cds_name,
            system_version=system_version,
            product_type=product_type,
            date_variable_prefix=date_variable_prefix,
            frequency=frequency,
            coord_names=coord_names,
            start_date_min=start_date_min,
            end_date_max=end_date_max,
            start_date=start_date,
            end_date=end_date,
            leadtime_max=leadtime_max,
        )

    @check_file_existence
    def _load_single_file(
        self, input_filepath: Path, filepath: Path, clobber: bool
    ) -> xr.Dataset:
        return _read_in_ensemble_and_perturbed_datasets(
            filepath=input_filepath
        )


def _read_in_ensemble_and_perturbed_datasets(filepath: Path):
    """Read in forecast and reforecast data.

    The GloFAS forecast and reforecast data GRIB files contain two
    separate datasets: the control member, generated from the most accurate
    estimate of current conditions, and the perturbed forecast, which
    contains N ensemble members created by perturbing the control forecast.

    This function reads in both datasets and creates an N+1 (perturbed
    + control) ensemble.
    See `this paper <https://hess.copernicus.org/preprints/hess-2020-532/>`__
    for more details.
    """
    ds_list = []
    for data_type in ["cf", "pf"]:
        ds = xr.load_dataset(
            filepath,
            engine="cfgrib",
            backend_kwargs={
                "indexpath": "",
                "filter_by_keys": {"dataType": data_type},
            },
        )
        # Extra processing require for control forecast
        if data_type == "cf":
            ds = ds.expand_dims(dim="number")
        ds_list.append(ds)
    return xr.combine_by_coords(ds_list, combine_attrs="drop_conflicts")


class GlofasForecast(_GlofasForecastBase):
    """
    Class for downloading and processing GloFAS forecast data.

    The GloFAS forecast dataset is a global raster presenting river
    discharge forecast from 26 May 2021 until present day (updated daily), see
    `this paper <https://hess.copernicus.org/preprints/hess-2020-532/>`__
    for more details. While CDS does have version 3 pre-release data
    from 2020-2021,
    we understand that there were some small issues that were fixed
    in the final version, so at this point in time this module
    does not support downloading the pre-release data.

    This class downloads the raw raster data
    `from CDS
    <https://cds.climate.copernicus.eu/cdsapp#!/dataset/cems-glofas-forecast?tab=overview>`__,
    and processes it from a raster to a datasets of reporting points from the
    `GloFAS interface
    <https://www.globalfloods.eu/glofas-forecasting/>`_.
    Due to the CDS request size limits, separate files are downloaded per
    day (that contain all requested lead times).

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    geo_bounding_box: GeoBoundingBox
        The bounding coordinates of the area that should be included
    leadtime_max: int
        The maximum desired lead time D in days. All forecast data for lead
        times 1 to D days are downloaded
    start_date : Union[date, str], default: date(year=2021, month=5, day=26)
        The starting date for the dataset. If left blank, defaults to the
        earliest available date
    end_date : Union[date, str], default: date.today()
        The ending date for the dataset. If left blank, defaults to
        the current date
    Examples
    --------
    Download, process and load GloFAS forecast data for the past month,
    for a lead time of 15 days.

    >>> from datetime import date
    >>> from aatoolbox import create_country_config, CodAB, GeoBoundingBox,
    ... GlofasForecast
    >>>
    >>> country_config = create_country_config(iso3="npl")
    >>> codab = CodAB(country_config=country_config)
    >>> codab.download()
    >>> admin_npl = codab.load()
    >>> geo_bounding_box = GeoBoundingBox.from_shape(admin_npl)
    >>>
    >>> glofas_forecast = GlofasForecast(
    ...     country_config=country_config,
    ...     geo_bounding_box=geo_bounding_box,
    ...     leadtime_max=15,
    ...     end_date=date(year=2022, month=10, day=22),
    ...     start_date=date(year=2022, month=9, day=22)
    ... )
    >>> glofas_forecast.download()
    >>> glofas_forecast.process()
    >>>
    >>> npl_glofas_forecast_reporting_points = glofas_forecast.load()
    """

    def __init__(
        self,
        country_config: CountryConfig,
        geo_bounding_box: GeoBoundingBox,
        leadtime_max: int,
        start_date: Union[date, str] = None,
        end_date: Union[date, str] = None,
    ):
        super().__init__(
            country_config=country_config,
            geo_bounding_box=geo_bounding_box,
            start_date=start_date,
            end_date=end_date,
            cds_name="cems-glofas-forecast",
            system_version="operational",
            product_type=["control_forecast", "ensemble_perturbed_forecasts"],
            date_variable_prefix="",
            frequency=rrule.DAILY,
            coord_names=["number", "step"],
            leadtime_max=leadtime_max,
            start_date_min=date(year=2021, month=5, day=26),
        )

    @staticmethod
    def _preprocess_load(ds: xr.Dataset) -> xr.Dataset:
        return ds.expand_dims("time")


class GlofasReforecast(_GlofasForecastBase):
    """
    Class for downloading and processing GloFAS reforecast data.

    The GloFAS reforecast dataset is a global raster presenting river
    discharnge forecasted from 1999 until 2018, see
    `this paper <https://hess.copernicus.org/preprints/hess-2020-532/>`_
    for more details.

    This class downloads the raw raster data
    `from CDS
    <https://cds.climate.copernicus.eu/cdsapp#!/dataset/cems-glofas-reforecast?tab=overview>`__,
    and processes it from a raster to a datasets of reporting points from the
    `GloFAS interface
    <https://www.globalfloods.eu/glofas-forecasting/>`_.
     Due to the CDS request size limits, separate files are downloaded per
     month (that contain all requested lead times).

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    geo_bounding_box: GeoBoundingBox
        The bounding coordinates of the area that should be included
    leadtime_max: int
        The maximum desired lead time D in days. All forecast data for lead
        times 1 to D days are downloaded
    start_date : Union[date, str], default: date(year=1999, month=1, day=1)
        The starting date for the dataset. If left blank, defaults to the
        earliest available date
    end_date : Union[date, str], default: date(year=2018, month=12, day=31)
        The ending date for the dataset. If left blank, defaults to the
        last available date

    Examples
    --------
    Download, process and load all available GloFAS reforecast data
    for a lead time of 15 days.

    >>> from aatoolbox import create_country_config, CodAB, GeoBoundingBox,
    ... GlofasReforecast
    >>>
    >>> country_config = create_country_config(iso3="npl")
    >>> codab = CodAB(country_config=country_config)
    >>> codab.download()
    >>> admin_npl = codab.load()
    >>> geo_bounding_box = GeoBoundingBox.from_shape(admin_npl)
    >>>
    >>> glofas_reforecast = GlofasReforecast(
    ...     country_config=country_config,
    ...     geo_bounding_box=geo_bounding_box,
    ...     leadtime_max=15
    ... )
    >>> glofas_reforecast.download()
    >>> glofas_reforecast.process()
    >>>
    >>> npl_glofas_reforecast_reporting_points = glofas_reforecast.load()
    """

    def __init__(
        self,
        country_config: CountryConfig,
        geo_bounding_box: GeoBoundingBox,
        leadtime_max: int,
        start_date: Union[date, str] = None,
        end_date: Union[date, str] = None,
    ):
        super().__init__(
            country_config=country_config,
            geo_bounding_box=geo_bounding_box,
            start_date=start_date,
            end_date=end_date,
            cds_name="cems-glofas-reforecast",
            system_version="version_3_1",
            product_type=[
                "control_reforecast",
                "ensemble_perturbed_reforecasts",
            ],
            date_variable_prefix="h",
            frequency=rrule.MONTHLY,
            coord_names=["number", "time", "step"],
            leadtime_max=leadtime_max,
            start_date_min=date(year=1999, month=1, day=1),
            end_date_max=date(year=2018, month=12, day=31),
        )



"""
FEWS NET processing.

Download and save the data provided by FEWS NET as provided
on <https://fews.net/>.

FEWS NET is only available in a set of countries.
Check their website to see which countries are included.
"""

import datetime
import logging
import zipfile
from enum import Enum
from pathlib import Path
from tempfile import TemporaryDirectory

import geopandas as gpd
from hdx.location.country import Country

from aatoolbox.datasources.datasource import DataSource
from aatoolbox.utils.check_file_existence import check_file_existence
from aatoolbox.utils.io import download_url, unzip

logger = logging.getLogger(__name__)
_BASE_URL_COUNTRY = (
    "https://fdw.fews.net/api/ipcpackage/"
    "?country_code={iso2}&collection_date={YYYY}-{MM}-01"
)
_BASE_URL_REGION = (
    "https://fews.net/data_portal_download/download"
    "?data_file_path=http://shapefiles.fews.net.s3.amazonaws.com/"
    "HFIC/{region_code}/{region_name}{YYYY}{MM}.zip"
)


# Use Enum such that it can function as type-checking
# as well as check if user-inputed string is valid
class ValidProjectionPeriods(Enum):
    """
    Define Enum for Valid Projection Periods.

    Use Enum such that it can function as type-checking
    as well as check if user-inputed string is valid
    """

    CS = "CS"
    ML1 = "ML1"
    ML2 = "ML2"


class FewsNet(DataSource):
    """
    Base class to retrieve FewsNet data.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    """

    def __init__(
        self,
        country_config,
    ):
        super().__init__(
            country_config=country_config,
            datasource_base_dir="fewsnet",
            is_public=True,
            # FN data can be regional, and thus we save the raw data
            # in "glb" instead of in the iso3 folder, even though part of it is
            # at country level
            is_global_raw=True,
            is_global_processed=False,
            config_datasource_name="fewsnet",
        )

        self._iso2 = Country.get_iso2_from_iso3(self._country_config.iso3)
        if self._iso2 is None:
            raise KeyError(
                "No ISO2 found for the given ISO3. Check your ISO3, currently:"
                f" {self._country_config.iso3}."
            )

    # mypy will give error Signature of "download" incompatible with supertype
    # "DataSource" due to `pub_year` and `pub_month` not being an arg in
    # `DataSource`. This is however valid so ignore mypy
    def download(  # type: ignore
        self,
        pub_year: int,
        pub_month: int,
        clobber: bool = False,
    ) -> Path:
        """
        Retrieve the raw FEWS NET data.

        Depending on the region and date, this data is published per region or
        per country. This function retrieves the country data
        if it exists, and else the regional data for `pub_year`-`pub_month`.

        Parameters
        ----------
        pub_year: int
            publication year of the data that should be downloaded
        pub_month: int
            publication month of the data that should be downloaded. This
            commonly refers to the month of the Current Situation period
        clobber : bool, default = False
            If True, overwrites existing raw files

        Returns
        -------
        Path to the downloaded file.

        Examples
        --------
        >>> from aatoolbox import create_country_config, FewsNet
        >>> # Download FEWS NET data for ETH published in 2021-06
        >>> country_config = create_country_config(iso3="eth")
        >>> fewsnet = FewsNet(country_config=country_config)
        >>> eth_fn_202106_path = fewsnet.download(pub_year=2021,pub_month=6)
        """
        self._check_date_validity(pub_year=pub_year, pub_month=pub_month)
        pub_month_str = self._get_pub_month_str(pub_month)
        # we prefer the country data as this more nicely structured
        # thus first check if that is available
        try:
            return self._download_country(
                pub_year=pub_year,
                pub_month_str=pub_month_str,
                clobber=clobber,
            )
        except zipfile.BadZipFile:
            try:
                return self._download_region(
                    pub_year=pub_year,
                    pub_month_str=pub_month_str,
                    clobber=clobber,
                )
            except zipfile.BadZipFile as err:
                raise RuntimeError(
                    "No country or regional data found for"
                    f" {pub_year}-{pub_month_str}. Check on the FEWS NET "
                    "website that data for your given date and country/region "
                    "exists."
                ) from err

    def process(self, *args, **kwargs):
        """
        Process FEWS NET data.

        Method not implemented.
        """
        logger.info("`process()` method not yet implemented for FEWS NET.")

    def load(  # type: ignore
        self,
        pub_year: int,
        pub_month: int,
        projection_period: ValidProjectionPeriods,
    ) -> gpd.GeoDataFrame:
        """
        Load FEWS NET data.

        For the given `pub_year`, `pub_month` and `projection_period`.

        Parameters
        ----------
        pub_year: int
            publication year of the data that should be loaded
        pub_month: int
            publication month of the data that should be loaded. This
            refers to the first month of the Current Situation period
        projection_period: str
            The projection period to be loaded. This should be CS, ML1, or ML2.
            Referring to Current Situation, near term projection, and medium
            term projection respectively.

        Returns
        -------
        Geopandas DataFrame with the specified data.

        Examples
        --------
        >>> from aatoolbox import create_country_config, FewsNet
        >>> # Load FEWS NET data for ETH published in 2021-06 of medium-term
        ... projection period (ML1)
        >>> country_config = create_country_config(iso3="eth")
        >>> fewsnet = FewsNet(country_config=country_config)
        >>> gdf_eth_fn_202106 = fewsnet.load(pub_year=2021,pub_month=6,
        ... projection_period = "ML1")
        """
        logger.warning(
            "load() now returns the raw data. In the future this "
            "will return processed data."
        )
        projperiods = [pp.value for pp in ValidProjectionPeriods]
        if projection_period not in projperiods:
            raise ValueError(
                f"{projection_period} is not a valid projection"
                f" period. It must be one of {', '.join(projperiods)}"
            )

        self._check_date_validity(pub_year=pub_year, pub_month=pub_month)
        pub_month_str = self._get_pub_month_str(pub_month)
        dir_path = self._find_raw_dir_date(
            pub_year=pub_year, pub_month_str=pub_month_str
        )
        file_path = self._get_raw_file_projection_period(
            dir_path=dir_path, projection_period=projection_period
        )
        return gpd.read_file(file_path)

    @staticmethod
    def _check_date_validity(pub_year: int, pub_month: int):
        try:
            pub_date = datetime.datetime(year=pub_year, month=pub_month, day=1)
        except ValueError as err:
            raise ValueError(
                f"The combination f pub_year-pub_month, "
                f"{pub_year}-{pub_month}, is not a valid date."
            ) from err
        if pub_date < datetime.datetime(year=2009, month=1, day=1):
            raise ValueError(
                f"FEWSNET publishes data since 2009, so adjust your pub_year "
                f"to be >=2009, currently {pub_year}"
            )
        elif pub_date > datetime.datetime.now():
            raise ValueError(
                "There is no data published in the future. The date should "
                "refer to the start month-year of the Current "
                "situation period."
            )

    @staticmethod
    def _get_pub_month_str(pub_month: int):
        return f"{pub_month:02d}"

    def _download_country(
        self, pub_year: int, pub_month_str: str, clobber: bool
    ) -> Path:
        """
        Download fewsnet data that covers the iso2 country.

        Returns
        -------
        country_data : Path
            if data found return the output_dir, else return None
        """
        url_country_date = _BASE_URL_COUNTRY.format(
            iso2=self._iso2, YYYY=pub_year, MM=pub_month_str
        )

        return self._download(
            url=url_country_date,
            area=self._iso2,
            pub_year=pub_year,
            pub_month_str=pub_month_str,
            clobber=clobber,
        )

    def _download_region(
        self,
        pub_year: int,
        pub_month_str: str,
        clobber: bool,
    ) -> Path:
        """
        Download fewsnet data that covers the region the iso3 belongs to.

        Returns
        -------
        region_data : Path
            If region data exists, return the saved dir else return None
        """
        url_region_date = _BASE_URL_REGION.format(
            region_code=self._datasource_config.region_code,
            region_name=self._datasource_config.region_name,
            YYYY=pub_year,
            MM=pub_month_str,
        )

        return self._download(
            url=url_region_date,
            area=self._datasource_config.region_code,
            pub_year=pub_year,
            pub_month_str=pub_month_str,
            clobber=clobber,
        )

    def _download(
        self,
        url: str,
        area: str,
        pub_year: int,
        pub_month_str: str,
        clobber: bool,
    ) -> Path:
        """
        Define output names and call _download_zip.

        url: str
            URL the zip file is located
        area: str
            Identifier of which area the data covers. This is either the ISO2
            or the region code
        pub_year: int
            publication year of the data that should be downloaded
        pub_month: str
            publication month of the data that should be downloaded. This
            commonly refers to the month of the Current Situation period
        """
        # filenames have upper iso2/regioncode, so use that for dirs as well
        output_dir = self._get_raw_dir_date(
            area=area, pub_year=pub_year, pub_month_str=pub_month_str
        )
        return self._download_zip(
            filepath=output_dir,
            zip_filename=self._get_zip_filename(
                area=area, pub_year=pub_year, pub_month_str=pub_month_str
            ),
            url=url,
            clobber=clobber,
        )

    def _get_raw_dir_date(self, area: str, pub_year: int, pub_month_str: str):
        return self._raw_base_dir / f"{area}_{pub_year}{pub_month_str}"

    @staticmethod
    def _get_zip_filename(area, pub_year, pub_month_str):
        return f"{area}{pub_year}{pub_month_str}.zip"

    @staticmethod
    @check_file_existence
    def _download_zip(
        filepath: Path, zip_filename: str, url: str, clobber: bool
    ) -> Path:
        """
        Download and unzip the file at the url.

        Parameters
        ----------
        zip_filename : str
            name of the zipfile
        url : str
            url that contains the zip file to be downloaded

        Returns
        -------
        output_dir : Path
            None if no valid file, else output_dir

        """
        # create tempdir to write zipfile to
        with TemporaryDirectory() as temp_dir:
            zip_path = Path(temp_dir) / zip_filename
            download_url(url=url, save_path=zip_path)
            logger.info(f"Downloaded {url} to {zip_path}")

            try:
                unzip(zip_file_path=zip_path, save_dir=filepath)
                logger.debug(f"Unzipped to {filepath}")

            except zipfile.BadZipFile as err:
                # indicates that the url returned something that wasn't a
                # zip, happens often and indicates data for the given
                # country - year-month is not available

                raise zipfile.BadZipFile(
                    f"No zip data returned from url {url} "
                    f"check that the area and year-month publication exist."
                ) from err

        return filepath

    def _find_raw_dir_date(self, pub_year: int, pub_month_str: str):
        """
        Check if a dir exists for the given `pub_year`-`pub_month`.

        Should either cover the iso2 or region.
        If exists, returns the dir path.
        """
        country_dir = self._get_raw_dir_date(
            area=self._iso2, pub_year=pub_year, pub_month_str=pub_month_str
        )
        region_dir = self._get_raw_dir_date(
            area=self._datasource_config.region_code,
            pub_year=pub_year,
            pub_month_str=pub_month_str,
        )
        if country_dir.is_dir():
            return country_dir
        elif region_dir.is_dir():
            return region_dir

        raise FileNotFoundError(
            f"No data found for {pub_year}-{pub_month_str} covering "
            f"{self._country_config.iso3} "
            f"or {self._datasource_config.region_name}. "
            f"Please make sure the data exists and is downloaded"
        )

    @staticmethod
    def _get_raw_file_projection_period(
        dir_path: Path, projection_period: ValidProjectionPeriods
    ):
        file_path = dir_path / f"{dir_path.name}_{projection_period}.shp"
        if file_path.is_file():
            return file_path
        else:
            raise FileNotFoundError(
                f"File {file_path} not found. Make sure the projection "
                f"period {projection_period} exists for {dir_path.name}."
            )



"""Use HDX python API to download data."""
import logging
import shutil
import tempfile
from pathlib import Path

from hdx.api.configuration import Configuration
from hdx.data.dataset import Dataset

USER_AGENT = "aa-toolbox"

logger = logging.getLogger(__name__)
Configuration.create(
    hdx_site="prod", user_agent=USER_AGENT, hdx_read_only=True
)


def load_dataset_from_hdx(
    hdx_address: str, hdx_dataset_name: str, output_filepath: Path
) -> Path:
    """
    Use the HDX API to download a dataset based on the address and dataset ID.

    Parameters
    ----------
    hdx_address : str
        The page where the dataset resides on on HDX. Can be found
        by taking the portion of the url after ``data.humdata.org/dataset/``
    hdx_dataset_name : str
        Dataset name on HDX. Can be found by taking the filename as it
        appears on the dataset page.
    output_filepath : Path
        Target filepath for the dataset

    Returns
    -------
    The full path of the downloaded dataset

    """
    logger.info(f"Querying HDX API for dataset {hdx_address}")
    resources = Dataset.read_from_hdx(hdx_address).get_resources()
    logger.debug(f"Found the following resources: {resources}")
    for resource in resources:
        if resource["name"] == hdx_dataset_name:
            logger.info(f"Downloading dataset {hdx_dataset_name}")
            with tempfile.TemporaryDirectory() as tempdir:
                _, downloaded_filepath = resource.download(folder=tempdir)
                output_filepath.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(downloaded_filepath, output_filepath)
            logger.info(f"Saved to {output_filepath}")
            return Path(output_filepath)
    raise FileNotFoundError(
        f'Dataset with name "{hdx_dataset_name}" not found'
        f'at HDX address "{hdx_address}".'
    )



"""Tests for the FewsNet module."""
import zipfile

import pytest
from conftest import ISO2

from aatoolbox.config.countryconfig import FewsNetConfig
from aatoolbox.datasources.fewsnet.fewsnet import FewsNet

DATASOURCE_BASE_DIR = "fewsnet"
_PUB_YEAR = 2020
_PUB_MONTH = 7
_PUB_MONTH_STR = f"{7:02d}"


@pytest.fixture(autouse=True)
def mock_iso2(mocker, request):
    """Mock iso2 to iso3 conversion."""
    if request.node.get_closest_marker("nomockiso2"):
        return
    mocker.patch(
        "aatoolbox.datasources.fewsnet.fewsnet.Country.get_iso2_from_iso3",
        return_value=ISO2.upper(),
    )


@pytest.mark.nomockiso2
def test_no_iso2(mock_country_config):
    """
    Test that if no valid iso2 can be found, a keyerror is returned.

    Since we use a fake iso3, no iso2 can be found and thus should produce an
    error. For the other tests, we mock the iso3-to-iso2 conversion to prevent
    this error.
    """
    with pytest.raises(KeyError):
        FewsNet(country_config=mock_country_config)


def test_download_country(
    mock_aa_data_dir, mock_country_config, mock_download_call, mocker
):
    """Test that the correct country url and path is returned."""
    fewsnet = FewsNet(country_config=mock_country_config)
    url, output_path = mock_download_call(
        fewsnet_class=fewsnet,
        pub_year=_PUB_YEAR,
        pub_month=_PUB_MONTH,
        country_data=True,
        region_data=False,
    )
    assert (
        url == "https://fdw.fews.net/api/ipcpackage/"
        f"?country_code={ISO2.upper()}&collection_date={_PUB_YEAR}-"
        f"{_PUB_MONTH_STR}-01"
    )
    assert (
        output_path
        == mock_aa_data_dir
        / "public"
        / "raw"
        / "glb"
        / DATASOURCE_BASE_DIR
        / f"{ISO2.upper()}_{_PUB_YEAR}{_PUB_MONTH_STR}"
    )


def test_download_region(
    mock_aa_data_dir, mock_country_config, mock_download_call
):
    """Test that the correct region url and path is returned."""
    fewsnet = FewsNet(country_config=mock_country_config)
    url, output_path = mock_download_call(
        fewsnet_class=fewsnet,
        pub_year=_PUB_YEAR,
        pub_month=_PUB_MONTH,
        country_data=False,
        region_data=True,
    )

    assert (
        url == f"https://fews.net/data_portal_download/download?"
        "data_file_path=http://shapefiles.fews.net.s3.amazonaws.com/"
        f"HFIC/EA/east-africa{_PUB_YEAR}{_PUB_MONTH_STR}.zip"
    )
    assert (
        output_path
        == mock_aa_data_dir
        / "public"
        / "raw"
        / "glb"
        / DATASOURCE_BASE_DIR
        / f"EA_{_PUB_YEAR}{_PUB_MONTH_STR}"
    )


def test_download_nodata(mock_country_config, mock_download_call):
    """Test that RuntimeError is returned when no data exists."""
    with pytest.raises(RuntimeError) as e:
        fewsnet = FewsNet(country_config=mock_country_config)
        url, output_path = mock_download_call(
            fewsnet_class=fewsnet,
            pub_year=_PUB_YEAR,
            pub_month=_PUB_MONTH,
            country_data=False,
            region_data=False,
        )
        assert output_path is None
    assert (
        f"No country or regional data found for {_PUB_YEAR}-{_PUB_MONTH_STR}"
        in str(e.value)
    )


def test_invalid_region_name():
    """Test raised error when too high admin level requested."""
    with pytest.raises(ValueError):
        FewsNetConfig.regionname_valid(
            v="supereast-africa",
            values={"region_name_code_mapping": {"east-africa": "EA"}},
        )


def test_date_valid(mock_country_config):
    """Test error when input date is not valid."""
    fewsnet = FewsNet(country_config=mock_country_config)
    with pytest.raises(ValueError):
        fewsnet._check_date_validity(pub_year=2000, pub_month=13)
        fewsnet._check_date_validity(pub_year=2000, pub_month=12)
        fewsnet._check_date_validity(pub_year=2100, pub_month=12)


@pytest.fixture
def mock_download_call(mock_fake_url, mock_countryregion):
    """Mock call to download."""
    url_mock = mock_fake_url

    def _get_country_mock(
        fewsnet_class, pub_year, pub_month, country_data, region_data
    ):
        mock_countryregion(country_data, region_data)

        output_path = fewsnet_class.download(
            pub_year=pub_year, pub_month=pub_month
        )

        _, kwargs_download_url = url_mock.call_args
        url = kwargs_download_url["url"]

        return url, output_path

    return _get_country_mock


@pytest.fixture
def mock_fake_url(mocker):
    """Mock url and unzip call."""
    fakedownloadurl = mocker.patch(
        "aatoolbox.datasources.fewsnet.fewsnet.download_url"
    )
    mocker.patch("aatoolbox.datasources.fewsnet.fewsnet.unzip")
    return fakedownloadurl


@pytest.fixture
def mock_countryregion(mocker):
    """Mock that no country and/or region data exists."""

    def _mock_countryregion(country_data: bool, region_data: bool):
        if country_data:
            return None
        else:
            if region_data:
                return mocker.patch(
                    "aatoolbox.datasources.fewsnet.fewsnet."
                    "FewsNet._download_country",
                    side_effect=zipfile.BadZipFile,
                )
            else:
                return mocker.patch(
                    "aatoolbox.datasources.fewsnet.fewsnet."
                    "FewsNet._download_country",
                    side_effect=zipfile.BadZipFile,
                ), mocker.patch(
                    "aatoolbox.datasources.fewsnet.fewsnet."
                    "FewsNet._download_region",
                    side_effect=zipfile.BadZipFile,
                )

    return _mock_countryregion


@pytest.fixture
def mock_gpd_read_file(mocker):
    """Mock GeoPandas file reading function."""
    return mocker.patch("aatoolbox.datasources.fewsnet.fewsnet.gpd.read_file")


def test_load(
    mock_country_config, mock_aa_data_dir, mock_gpd_read_file, mocker
):
    """Test loading of fewsnet data."""
    fewsnet = FewsNet(country_config=mock_country_config)
    mocker.patch("pathlib.Path.is_dir", return_value=True)
    mocker.patch("pathlib.Path.is_file", return_value=True)
    fewsnet.load(
        pub_year=_PUB_YEAR, pub_month=_PUB_MONTH, projection_period="CS"
    )
    mock_gpd_read_file.assert_has_calls(
        [
            mocker.call(
                mock_aa_data_dir
                / "public"
                / "raw"
                / "glb"
                / DATASOURCE_BASE_DIR
                / f"{ISO2.upper()}_{_PUB_YEAR}{_PUB_MONTH_STR}"
                / f"{ISO2.upper()}_{_PUB_YEAR}{_PUB_MONTH_STR}_CS.shp"
            ),
        ]
    )


def test_invalid_projection_period(mock_country_config):
    """Test that fails when projection_period is not one of the options."""
    fewsnet = FewsNet(country_config=mock_country_config)
    with pytest.raises(ValueError):
        fewsnet.load(
            pub_year=_PUB_YEAR, pub_month=_PUB_MONTH, projection_period="CA"
        )


def test_download_clobber(mock_country_config, mock_fake_url):
    """Test that download URL is not called if directory exists."""
    fewsnet = FewsNet(country_config=mock_country_config)
    filepath = fewsnet._get_raw_dir_date(
        area=fewsnet._iso2,
        pub_year=_PUB_YEAR,
        pub_month_str=f"{_PUB_MONTH:02d}",
    )
    filepath.mkdir(parents=True)
    fewsnet.download(pub_year=_PUB_YEAR, pub_month=_PUB_MONTH, clobber=False)
    mock_fake_url.assert_not_called()



"""Download and process GloFAS reanalysis river discharge data."""
import logging
from datetime import date
from pathlib import Path
from typing import Union

import xarray as xr
from dateutil import rrule

from aatoolbox.config.countryconfig import CountryConfig
from aatoolbox.datasources.glofas import glofas
from aatoolbox.utils.check_file_existence import check_file_existence
from aatoolbox.utils.geoboundingbox import GeoBoundingBox

logger = logging.getLogger(__name__)


class GlofasReanalysis(glofas.Glofas):
    """
    Class for downloading and processing GloFAS reanalysis data.

    The GloFAS reanalysis dataset is a global raster presenting river
    discharnge from 1979 until present day (updated daily), see
    `this paper <https://essd.copernicus.org/articles/12/2043/2020/>`_
    for more details.

    This class downloads the raw raster data
    `from CDS
    <https://cds.climate.copernicus.eu/cdsapp#!/dataset/cems-glofas-historical?tab=overview>`_,
    and processes it from a raster to a datasets of reporting points from the
    `GloFAS interface
    <https://www.globalfloods.eu/glofas-forecasting/>`_.
    Due to the CDS request size limits, separate files are downloaded per
    year.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    geo_bounding_box: GeoBoundingBox
        The bounding coordinates of the area that should be included
    start_date : Union[date, str], default: date(1979, 1, 1)
        The starting date for the dataset. If left blank, defaults to the
        earliest available date
    end_date : Union[date, str], default: date.today()
        The ending date for the dataset. If left blank, defaults to
        the current date
    Examples
    --------
    Download, process and load all historical GloFAS reanalysis data
    until the current date, set to Oct 22, 2022 for this example.

    >>> from datetime import date
    >>> from aatoolbox import create_country_config, CodAB, GeoBoundingBox,
    ... GlofasReanalysis
    >>>
    >>> country_config = create_country_config(iso3="bgd")
    >>> codab = CodAB(country_config=country_config)
    >>> codab.download()
    >>> admin_npl = codab.load()
    >>> geo_bounding_box = GeoBoundingBox.from_shape(admin_npl)
    >>>
    >>> glofas_reanalysis = GlofasReanalysis(
    ...     country_config=country_config,
    ...     geo_bounding_box=geo_bounding_box,
    ...     end_date=date(year=2022, month=10, day=22)
    ... )
    >>> glofas_reanalysis.download()
    >>> glofas_reanalysis.process()
    >>>
    >>> npl_glofas_reanalysis_reporting_points = glofas_reanalysis.load()
    """

    def __init__(
        self,
        country_config: CountryConfig,
        geo_bounding_box: GeoBoundingBox,
        start_date: Union[date, str] = None,
        end_date: Union[date, str] = None,
    ):
        super().__init__(
            country_config=country_config,
            geo_bounding_box=geo_bounding_box,
            cds_name="cems-glofas-historical",
            system_version="version_3_1",
            product_type="consolidated",
            date_variable_prefix="h",
            frequency=rrule.YEARLY,
            coord_names=["time"],
            start_date_min=date(year=1979, month=1, day=1),
            start_date=start_date,
            end_date=end_date,
        )

    @check_file_existence
    def _load_single_file(
        self, input_filepath: Path, filepath: Path, clobber: bool
    ) -> xr.Dataset:
        return xr.load_dataset(
            input_filepath, engine="cfgrib", backend_kwargs={"indexpath": ""}
        )



"""Fixtures for all pipeline-related tests."""
import pytest

from aatoolbox import GeoBoundingBox, create_custom_country_config

CONFIG_FILE = "tests/datasources/fake_config.yaml"
ISO3 = "abc"
ISO2 = "ab"


@pytest.fixture(autouse=True)
def mock_aa_data_dir(tmp_path_factory, mocker):
    """Mock out the AA_DATA_DIR environment variable."""
    mock_aa_data_dir_path = tmp_path_factory.mktemp(
        basename="test_aa_data_dir"
    )
    mocker.patch.dict(
        "aatoolbox.config.pathconfig.os.environ",
        {"AA_DATA_DIR": str(mock_aa_data_dir_path)},
    )
    return mock_aa_data_dir_path


@pytest.fixture
def mock_country_config():
    """Fixture for pipeline with test config params."""
    return create_custom_country_config(filepath=CONFIG_FILE)


@pytest.fixture
def geo_bounding_box():
    """Input GeoBoundingBox to use."""
    gbb = GeoBoundingBox(lat_max=1.0, lat_min=-2.2, lon_max=3.3, lon_min=-4.4)
    return gbb


def pytest_configure(config):
    """Create custom markers to add to tests."""
    config.addinivalue_line(
        "markers",
        (
            "nomockiso2: do not mock the get_iso2_from_iso3 function. "
            "Used for FEWS NET"
        ),
    )



"""Test dates module."""

from datetime import date, datetime

import pytest

from aatoolbox.utils import dates


def test_get_date():
    """Tests getting date from string."""
    desired_date = date(year=2018, month=3, day=14)
    desired_str = "2018-03-14"
    desired_datetime = datetime(year=2018, month=3, day=14)
    assert dates.get_date_from_user_input(desired_str) == desired_date
    assert dates.get_date_from_user_input(desired_date) == desired_date
    assert dates.get_date_from_user_input(desired_datetime) == desired_datetime


def test_get_date_value_error():
    """Tests getting date errors."""
    with pytest.raises(ValueError):
        dates.get_date_from_user_input("2013-3-14")
    with pytest.raises(TypeError):
        dates.get_date_from_user_input(None)


def test_get_dekadal_date():
    """Tests get dekadal date."""
    desired_dekad = (2013, 8)
    desired_str = "2013-03-14"
    assert dates.get_dekadal_date(desired_str) == desired_dekad
    assert dates.get_dekadal_date(desired_dekad) == desired_dekad
    assert (
        dates.get_dekadal_date(date.fromisoformat(desired_str))
        == desired_dekad
    )
    assert (
        dates.get_dekadal_date(input_date=None, default_date=desired_dekad)
        == desired_dekad
    )


def test_get_dekadal_date_value_error():
    """Test that value error raised for bad inputs."""
    with pytest.raises(ValueError):
        dates.get_dekadal_date("2020-01-1")
    with pytest.raises(ValueError):
        dates.get_dekadal_date((20, 17))


def test_dekad_to_date():
    """Test conversion from dekad to date."""
    assert dates.dekad_to_date((2020, 1)) == date.fromisoformat("2020-01-01")
    assert dates.dekad_to_date((2016, 20)) == date.fromisoformat("2016-07-11")
    assert dates.dekad_to_date((2024, 36)) == date.fromisoformat("2024-12-21")


def test_date_to_dekad():
    """Test conversion from dekad to date."""
    assert dates.date_to_dekad(date.fromisoformat("2020-01-09")) == (2020, 1)
    assert dates.date_to_dekad(date.fromisoformat("2016-07-20")) == (2016, 20)
    assert dates.date_to_dekad(date.fromisoformat("2024-12-31")) == (2024, 36)


def test_compare_dekads():
    """Test comparing dekads."""
    dekad1 = (2019, 36)
    dekad2 = (2020, 1)
    assert dates.compare_dekads_gt(dekad2, dekad1)
    assert dates.compare_dekads_gte(dekad2, dekad1)
    assert dates.compare_dekads_lt(dekad1, dekad2)
    assert dates.compare_dekads_lte(dekad1, dekad2)
    assert dates.compare_dekads_gte(dekad2, dekad2)
    assert dates.compare_dekads_lte(dekad2, dekad2)


def test_expand_dekads():
    """Test expanding dekads."""
    dekad1 = (2019, 33)
    dekad2 = (2020, 3)
    dekads = dates.expand_dekads(dekad1, dekad2)
    assert len(dekads) == 7
    assert dekads[0] == dekad1
    assert dekads[6] == dekad2
    assert dekads[4] == (2020, 1)
    # error raised for reverse attempt
    with pytest.raises(ValueError):
        dates.expand_dekads(dekad2, dekad1)



"""Tests for GloFAS data download and processing."""
from datetime import date
from pathlib import Path
from unittest import mock

import pytest

from aatoolbox import (
    GeoBoundingBox,
    GlofasForecast,
    GlofasReanalysis,
    GlofasReforecast,
)


@pytest.fixture
def mock_retrieve(mocker):
    """Mock only the retrieve method of the Client."""
    mocker.patch(
        "aatoolbox.datasources.glofas.glofas.cdsapi.Client.__init__",
        return_value=None,
    )
    return mocker.patch(
        "aatoolbox.datasources.glofas.glofas.cdsapi.Client.retrieve"
    )


@pytest.fixture
def mock_result(mocker):
    """Mock the entire Result class."""
    # This is needed because the reply is changed dynamically
    # so very difficult to use a Mock object
    class MockResult:
        def __init__(self, *args, **kwargs):
            self.reply = {"state": "completed"}
            self.state = None

        def update(self):
            pass

        @mock.create_autospec
        def download(self, filepath):
            pass

    return mocker.patch(
        "aatoolbox.datasources.glofas.glofas.cdsapi.api.Result",
        return_value=MockResult(),
    )


@pytest.fixture
def geo_bounding_box():
    """Input GeoBoundingBox to use."""
    gbb = GeoBoundingBox(lat_max=1.0, lat_min=-2.2, lon_max=3.3, lon_min=-4.4)
    return gbb


def test_reanalysis_download(
    mock_country_config,
    mock_aa_data_dir,
    mock_retrieve,
    mock_result,
    geo_bounding_box,
):
    """
    Test GloFAS reanalysis download.

    Test that the query generated by the download method of GlofasReanlysis
    with default parameters is as expected
    """
    glofas_reanalysis = GlofasReanalysis(
        country_config=mock_country_config,
        geo_bounding_box=geo_bounding_box,
        start_date=date(year=2021, month=1, day=1),
        end_date=date(year=2021, month=12, day=31),
    )
    glofas_reanalysis.download()
    expected_retrieve_args = {
        "name": "cems-glofas-historical",
        "request": {
            "variable": "river_discharge_in_the_last_24_hours",
            "format": "grib",
            "product_type": "consolidated",
            "system_version": "version_3_1",
            "hydrological_model": "lisflood",
            "hyear": "2021",
            "hmonth": [str(x + 1).zfill(2) for x in range(12)],
            "hday": [str(x + 1).zfill(2) for x in range(31)],
            "area": [1.05, -4.45, -2.25, 3.35],
        },
    }
    expected_result_path = Path(
        f"{mock_aa_data_dir}/public/raw/{mock_country_config.iso3}"
        f"/glofas/cems-glofas-historical/"
        f"{mock_country_config.iso3}_"
        f"cems-glofas-historical_2021_Np1d05Sm2d25Ep3d35Wm4d45.grib"
    )
    mock_retrieve.assert_called_with(**expected_retrieve_args)
    mock_result.return_value.download.assert_called_with(
        mock.ANY, expected_result_path
    )


def test_forecast_download(
    mock_country_config,
    mock_aa_data_dir,
    mock_retrieve,
    mock_result,
    geo_bounding_box,
):
    """
    Test GloFAS forecast download.

    Test that the query generated by the download method of GlofasForecast
    with default parameters is as expected
    """
    glofas_forecast = GlofasForecast(
        country_config=mock_country_config,
        geo_bounding_box=geo_bounding_box,
        leadtime_max=3,
        start_date=date(year=2022, month=1, day=1),
        end_date=date(year=2022, month=1, day=1),
    )
    glofas_forecast.download()
    expected_retrieve_args = {
        "name": "cems-glofas-forecast",
        "request": {
            "variable": "river_discharge_in_the_last_24_hours",
            "format": "grib",
            "product_type": [
                "control_forecast",
                "ensemble_perturbed_forecasts",
            ],
            "system_version": "operational",
            "hydrological_model": "lisflood",
            "year": "2022",
            "month": "01",
            "day": "01",
            "area": [1.05, -4.45, -2.25, 3.35],
            "leadtime_hour": ["24", "48", "72"],
        },
    }

    expected_result_path = Path(
        f"{mock_aa_data_dir}/public/raw/{mock_country_config.iso3}/"
        f"glofas/cems-glofas-forecast/"
        f"{mock_country_config.iso3}_"
        f"cems-glofas-forecast_2022-01-01_ltmax03d_Np1d05Sm2d25Ep3d35Wm4d45"
        f".grib"
    )
    mock_retrieve.assert_called_with(**expected_retrieve_args)
    mock_result.return_value.download.assert_called_with(
        mock.ANY, expected_result_path
    )


def test_reforecast_download(
    mock_country_config,
    mock_aa_data_dir,
    mock_retrieve,
    mock_result,
    geo_bounding_box,
):
    """
    Test GloFAS reforecast download.

    Test that the query generated by the download method of
    GlofasReforecast with default parameters is as expected
    """
    glofas_reforecast = GlofasReforecast(
        country_config=mock_country_config,
        geo_bounding_box=geo_bounding_box,
        leadtime_max=3,
        start_date=date(year=2018, month=1, day=1),
        end_date=date(year=2018, month=1, day=31),
    )
    glofas_reforecast.download()
    expected_retrieve_args = {
        "name": "cems-glofas-reforecast",
        "request": {
            "variable": "river_discharge_in_the_last_24_hours",
            "format": "grib",
            "product_type": [
                "control_reforecast",
                "ensemble_perturbed_reforecasts",
            ],
            "system_version": "version_3_1",
            "hydrological_model": "lisflood",
            "hyear": "2018",
            "hmonth": "01",
            "hday": [str(x + 1).zfill(2) for x in range(31)],
            "area": [1.05, -4.45, -2.25, 3.35],
            "leadtime_hour": ["24", "48", "72"],
        },
    }
    expected_result_path = Path(
        f"{mock_aa_data_dir}/public/raw/{mock_country_config.iso3}/"
        f"glofas/cems-glofas-reforecast/"
        f"{mock_country_config.iso3}_"
        f"cems-glofas-reforecast_2018-01_ltmax03d_Np1d05Sm2d25Ep3d35Wm4d45"
        f".grib"
    )
    mock_retrieve.assert_called_with(**expected_retrieve_args)
    mock_result.return_value.download.assert_called_with(
        mock.ANY, expected_result_path
    )



"""Tests for the IRI module."""
import cftime
import numpy as np
import pytest
import requests
import xarray as xr
from xarray.coding.cftimeindex import CFTimeIndex

from aatoolbox import GeoBoundingBox, IriForecastDominant, IriForecastProb

DATASOURCE_BASE_DIR = "iri"
FAKE_IRI_AUTH = "FAKE_IRI_AUTH"


@pytest.fixture
def mock_iri(mock_country_config):
    """Create IRI class with mock country config."""
    geo_bounding_box = GeoBoundingBox(
        lat_max=6, lat_min=3.2, lon_max=2, lon_min=-3
    )

    def _mock_iri(prob_forecast: bool = True):
        if prob_forecast:
            iri = IriForecastProb(
                country_config=mock_country_config,
                geo_bounding_box=geo_bounding_box,
            )
        else:
            iri = IriForecastDominant(
                country_config=mock_country_config,
                geo_bounding_box=geo_bounding_box,
            )
        return iri

    return _mock_iri


@pytest.fixture
def mock_download(mocker, mock_iri):
    """
    Call download with mocked _download.

    `forecast_type` is the type of forecast to
    test, can be either 'prob' or 'dominant'.
    """
    download_mock = mocker.patch(
        (
            "aatoolbox.datasources.iri.iri_seasonal_forecast"
            "._IriForecast._download"
        )
    )

    mocker.patch.dict(
        "aatoolbox.datasources.iri.iri_seasonal_forecast.os.environ",
        {"IRI_AUTH": FAKE_IRI_AUTH},
    )

    def _mock_download(prob_forecast: bool):
        iri = mock_iri(prob_forecast=prob_forecast)
        iri.download()
        _, kwargs_download = download_mock.call_args
        url = kwargs_download["url"]
        filepath = kwargs_download["filepath"]
        return url, filepath

    return _mock_download


def test_download_call_prob(
    mock_download, mock_aa_data_dir, mock_country_config
):
    """Test download for tercile probability forecast."""
    url, filepath = mock_download(prob_forecast=True)
    assert url == (
        "https://iridl.ldeo.columbia.edu/SOURCES/.IRI/.FD/"
        ".NMME_Seasonal_Forecast/"
        ".Precipitation_ELR/.prob/X/%28-3.0%29%282.0%29RANGEEDGES/"
        "Y/%286.0%29%283.0%29RANGEEDGES/data.nc"
    )

    assert filepath == (
        mock_aa_data_dir
        / f"private/raw/{mock_country_config.iso3}/{DATASOURCE_BASE_DIR}/"
        f"abc_iri_forecast_seasonal_precipitation_"
        f"tercile_prob_Np6Sp3Ep2Wm3.nc"
    )


def test_download_call_dominant(
    mock_download, mock_aa_data_dir, mock_country_config
):
    """Test download for dominant tercile forecast."""
    url, filepath = mock_download(prob_forecast=False)
    assert url == (
        "https://iridl.ldeo.columbia.edu/SOURCES/.IRI/.FD/"
        ".NMME_Seasonal_Forecast/"
        ".Precipitation_ELR/.dominant/X/%28-3.0%29%282.0%29RANGEEDGES/"
        "Y/%286.0%29%283.0%29RANGEEDGES/data.nc"
    )

    assert filepath == (
        mock_aa_data_dir
        / f"private/raw/{mock_country_config.iso3}/{DATASOURCE_BASE_DIR}/"
        f"abc_iri_forecast_seasonal_"
        f"precipitation_tercile_dominant_Np6Sp3Ep2Wm3.nc"
    )


@pytest.fixture
def mock_requests(mocker):
    """Mock requests in the download function."""
    requests_mock = mocker.patch(
        "aatoolbox.datasources.iri.iri_seasonal_forecast.requests.get"
    )
    mocker.patch.dict(
        "aatoolbox.datasources.iri.iri_seasonal_forecast.os.environ",
        {"IRI_AUTH": FAKE_IRI_AUTH},
    )

    return requests_mock


def test_download_wrong_auth(
    mock_iri, mock_requests, mock_aa_data_dir, mock_country_config
):
    """Check that wrong download headers raise an error."""
    iri = mock_iri(prob_forecast=True)
    with pytest.raises(requests.RequestException):
        iri.download()


def test_process(mocker, mock_iri, mock_aa_data_dir, mock_country_config):
    """Test process for IRI forecast."""
    ds = xr.DataArray(
        np.reshape(a=np.arange(16), newshape=(2, 2, 2, 2)),
        dims=("L", "X", "Y", "F"),
        coords={
            "L": [1, 2],
            "X": [2, -3],
            "Y": [97, 90],
            "F": [685.5, 686.5],
        },
    ).to_dataset(name="prob")

    ds["F"].attrs["calendar"] = "360"
    ds["F"].attrs["units"] = "months since 1960-01-01"

    iri = mock_iri()

    # TODO: now created `load_raw` to be able to mock but would like
    #  to do it from xr.load_dataset directly
    mocker.patch(
        "aatoolbox.datasources.iri.iri_seasonal_forecast."
        "_IriForecast._load_raw",
        return_value=ds,
    )

    processed_path = iri.process()
    assert processed_path == (
        mock_aa_data_dir / f"private/processed/{mock_country_config.iso3}/"
        f"{DATASOURCE_BASE_DIR}/{mock_country_config.iso3}_"
        f"iri_forecast_seasonal_precipitation_"
        f"tercile_prob_Np6Sp3Ep2Wm3.nc"
    )

    da_processed = xr.load_dataset(processed_path)
    expected_f = CFTimeIndex(
        [
            cftime.datetime(year=2017, month=2, day=16, calendar="360_day"),
            cftime.datetime(year=2017, month=3, day=16, calendar="360_day"),
        ]
    )

    assert np.array_equal(da_processed.X.values, [2, -3])
    assert np.array_equal(da_processed.Y.values, [97, 90])
    assert da_processed.get_index("F").equals(expected_f)
    assert np.array_equal(da_processed.prob.values, ds.prob.values)


def test_process_if_download_not_called(mock_iri):
    """Test that correct error message raised."""
    iri = mock_iri()
    # Make sure file doesn't exist
    if iri._get_raw_path().exists():
        iri._get_raw_path().unlink()
    with pytest.raises(FileNotFoundError) as excinfo:
        iri.process()
    assert (
        "Make sure that you have already called the 'download' method"
        in str(excinfo.value)
    )


@pytest.fixture
def mock_xr_load_dataset(mocker):
    """Mock GeoPandas file reading function."""
    return mocker.patch(
        "aatoolbox.datasources.iri.iri_seasonal_forecast.xr.load_dataset"
    )


def test_iri_load(
    mocker,
    mock_xr_load_dataset,
    mock_iri,
    mock_aa_data_dir,
    mock_country_config,
):
    """Test that load_codab calls the HDX API to download."""
    mocker.patch(
        (
            "aatoolbox.datasources.iri.iri_seasonal_forecast"
            "._IriForecast._download"
        )
    )

    iri = mock_iri()
    iri.load()
    mock_xr_load_dataset.assert_has_calls(
        [
            mocker.call(
                mock_aa_data_dir
                / f"private/processed/{mock_country_config.iso3}/"
                f"{DATASOURCE_BASE_DIR}/{mock_country_config.iso3}"
                f"_iri_forecast_seasonal_precipitation_"
                f"tercile_prob_Np6Sp3Ep2Wm3.nc"
            ),
        ]
    )


def test_load_if_process_not_called(mock_iri):
    """Test that correct error message raised."""
    iri = mock_iri()
    # Make sure file doesn't exist
    if iri._get_processed_path().exists():
        iri._get_processed_path().unlink()
    with pytest.raises(FileNotFoundError) as excinfo:
        iri.load()
    assert (
        "Make sure that you have already called the 'process' method"
        in str(excinfo.value)
    )



"""
Functionality to retrieve and modify boundary coordinates.

It is possible to create an ``GeoBoundingBox`` object either from
lat_max, lat_min, lon_max, lon_min coordinates,
or from a shapefile that has been read in with geopandas.
"""
import logging
from decimal import Decimal, getcontext
from typing import Union

import geopandas as gpd
import numpy as np

logger = logging.getLogger(__name__)
getcontext().prec = 5  # Assuming we won't need higher than this!


class GeoBoundingBox:
    """Create an object containing the bounds of an area.

    Standard geographic coordinate system is used where latitude runs
    from -90 to 90 degrees, and latitude from -180 to 180.
    North must always be greater than south, and east greater
    than west.

    Parameters
    ----------
    lat_max : float
        The northern latitude boundary of the area (degrees).
        The value must be between -90 and 90, and greater than or equal to the
        southern boundary.
    lat_min : float
        The southern latitude boundary of the area (degrees).
        The value must be between -90 and 90, and less than or equal to the
        northern boundary.
    lon_max : float
        The easternmost longitude boundary of the area (degrees).
        The value must be between -180 and 180, and greater than or equal to
        the western boundary.
    lon_min : float
        The westernmost longitude boundary of the area (degrees).
        The value must be between -180 and 180, and less than or equal to the
        eastern boundary.
    """

    def __init__(
        self, lat_max: float, lat_min: float, lon_max: float, lon_min: float
    ):
        self.lat_max = lat_max
        self.lat_min = lat_min
        self.lon_max = lon_max
        self.lon_min = lon_min

    @property
    def lat_max(self) -> float:
        """Get the northern latitude boundary of the area (degrees)."""
        return float(self._lat_max)

    @lat_max.setter
    def lat_max(self, lat_max):
        _check_latitude(lat_max)
        self._lat_max = Decimal(lat_max)

    @property
    def lat_min(self) -> float:
        """Get the southern latitude boundary of the area (degrees)."""
        return float(self._lat_min)

    @lat_min.setter
    def lat_min(self, lat_min):
        _check_latitude(lat_min)
        if not lat_min <= self.lat_max:
            raise AttributeError(
                "The maximum latitude must be greater than or equal to"
                "the minimum latitude"
            )
        self._lat_min = Decimal(lat_min)

    @property
    def lon_max(self) -> float:
        """Get the eastern longitude boundary of the area (degrees)."""
        return float(self._lon_max)

    @lon_max.setter
    def lon_max(self, lon_max):
        _check_longitude(lon_max)
        self._lon_max = Decimal(lon_max)

    @property
    def lon_min(self) -> float:
        """Get the western longitude boundary of the area (degrees)."""
        return float(self._lon_min)

    @lon_min.setter
    def lon_min(self, lon_min):
        _check_longitude(lon_min)
        if not lon_min <= self.lon_max:
            raise AttributeError(
                "The maximum longitude must be greater than or equal to "
                "the minimum longitude"
            )
        self._lon_min = Decimal(lon_min)

    def __repr__(self):
        """Print bounding box string."""
        return (
            f"N: {self.lat_max}\nS: {self.lat_min}\n"
            f"E: {self.lon_max}\nW: {self.lon_min}"
        )

    @classmethod
    def from_shape(
        cls, shape: Union[gpd.GeoSeries, gpd.GeoDataFrame]
    ) -> "GeoBoundingBox":
        """
        Create ``GeoBoundingBox`` from a geopandas object.

        Parameters
        ----------
        shape : geopandas.GeoSeries, geopandas.GeoDataFrame
            A shape whose bounds will be retrieved

        Returns
        -------
        ``GeoBoundingBox`` from the total bounds of the ``GeoDataFrame``

        Examples
        --------
        >>> import geopandas as gpd
        >>> df_admin_boundaries = gpd.read_file("admin0_boundaries.gpkg")
        >>> geobb = GeoBoundingBox.from_shape(df_admin_boundaries)
        """
        return cls(
            lat_max=shape.total_bounds[3],
            lat_min=shape.total_bounds[1],
            lon_max=shape.total_bounds[2],
            lon_min=shape.total_bounds[0],
        )

    def round_coords(
        self,
        offset_val: float = 0.0,
        round_val: Union[int, float] = 1,
    ) -> "GeoBoundingBox":
        """
        Round the bounding box coordinates.

        Rounding is always done outside the original bounding box,
        i.e. the resulting bounding box is always equal or larger
        than the original bounding box. Rounding can only be done once
        per instance.

        Parameters
        ----------
        offset_val : float, default = 0.0
            Offset the coordinates by this factor.
        round_val : int or float, default = 1
            Rounds to the nearest round_val. Can be an int for integer
            rounding or float for decimal rounding. If 1, round to integers.

        Returns
        -------
        ``GeoBoundingBox`` instance with rounded and offset coordinates
        """
        # TODO: add examples above
        new_coords = {}
        for direction in ["lat_max", "lon_min", "lat_min", "lon_max"]:
            coord = getattr(self, f"_{direction}")
            if direction in ("lat_max", "lon_max"):
                function = np.ceil.__call__  # needed for mypy
                offset_factor = Decimal(1)
            elif direction in ("lat_min", "lon_min"):
                function = np.floor.__call__  # needed for mypy
                offset_factor = Decimal(-1)
            rounded_coord = function(coord / Decimal(round_val)) * Decimal(
                round_val
            ) + offset_factor * Decimal(offset_val)
            new_coords[direction] = float(rounded_coord)
        return GeoBoundingBox(**new_coords)

    def get_filename_repr(self, p: int = 0) -> str:
        """
        Get succinct boundary representation for usage in filenames.

        Parameters
        ----------
        p : int, default = 0
            Precision, i.e. number of decimal places to round to. Default is
            0 for ints.

        Returns
        -------
        String containing N, S, E and W coordinates.
        """

        def _str_format(coord):
            """Add m indicating minus value and p indicating positive value."""
            if coord < 0:
                return f"m{abs(coord):.{p}f}"
            else:
                return f"p{coord:.{p}f}"

        return (
            f"N{_str_format(self.lat_max)}S{_str_format(self.lat_min)}"
            f"E{_str_format(self.lon_max)}W{_str_format(self.lon_min)}"
            # replace the decimal dot with a d for a better filename
            .replace(".", "d")
        )


def _check_latitude(value):
    if not -90 <= value <= 90:
        raise AttributeError("Latitude must range from -90 to 90 degrees")


def _check_longitude(value):
    if not -180 <= value <= 180:
        raise AttributeError("Longitude must range from -180 to 180 degrees")



"""Class to download and process USGS eMODIS NDVI data.

Download, process, and load eMODIS NDVI data published
in the `USGS FEWS NET data portal
<https://earlywarning.usgs.gov/fews>`_.

Warning: the MODIS sensor has been reported by USGS to
have degraded in quality since May 2022 (dekad 13), and
updates to this data source have stopped. This module
remains for users to have access to historic data but
recent data is unavailable and care should be used
analyzing any data since dekad 13 of 2022.
"""

# TODO: add progress bar
import functools
import logging
import re
from datetime import date, datetime
from io import BytesIO
from pathlib import Path
from typing import Generator, List, Optional, Tuple, Union
from urllib.error import HTTPError
from urllib.request import urlopen
from zipfile import ZipFile

import geopandas as gpd
import pandas as pd
import rioxarray  # noqa: F401
import xarray as xr
from rasterio.errors import RasterioIOError

import aatoolbox.utils.raster  # noqa: F401
from aatoolbox.config.countryconfig import CountryConfig
from aatoolbox.datasources.datasource import DataSource
from aatoolbox.utils.dates import (
    compare_dekads_gt,
    compare_dekads_lt,
    dekad_to_date,
    expand_dekads,
    get_dekadal_date,
)

logger = logging.getLogger(__name__)

_DATE_TYPE = Union[date, str, Tuple[int, int], None]
_EARLIEST_DATE = (2002, 19)

# USGS has reported degradation of USGS NDVI data
# from the below date and warnings should be used
_DEGRADATION_DATE = (2022, 13)


class _UsgsNdvi(DataSource):
    """Base class to retrieve USGS NDVI data.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    data_variable : str
        Data variable date
    data_variable_suffix : str
        Data variable file string
    data_variable_url : str
        URL string for data variable
    start_date : _DATE_TYPE, default = None
        Start date. Can be passed as a ``datetime.date``
        object or a data string in ISO8601 format, and
        the relevant dekad will be determined. Or pass
        directly as year-dekad tuple, e.g. (2020, 1).
        If ``None``, ``start_date`` is set to earliest
        date with data: 2002, dekad 19.
    end_date : _DATE_TYPE, default = None
        End date. Can be passed as a ``datetime.date``
        object and the relevant dekad will be determined,
        as a date string in ISO8601 format, or as a
        year-dekad tuple, i.e. (2020, 1). If ``None``,
        ``end_date`` is set to ``date.today()``.
    """

    def __init__(
        self,
        country_config: CountryConfig,
        data_variable: str,
        data_variable_suffix: str,
        data_variable_url: str,
        start_date: _DATE_TYPE = None,
        end_date: _DATE_TYPE = None,
    ):
        super().__init__(
            country_config=country_config,
            datasource_base_dir="usgs_ndvi",
            is_public=True,
            is_global_raw=True,
            config_datasource_name="usgs_ndvi",
        )

        # set data variable
        self._data_variable = data_variable
        self._data_variable_url = data_variable_url
        self._data_variable_suffix = data_variable_suffix

        # set dates for data download and processing
        self._start_date = get_dekadal_date(
            input_date=start_date, default_date=_EARLIEST_DATE
        )

        self._end_date = get_dekadal_date(
            input_date=end_date, default_date=date.today()
        )

        if compare_dekads_gt(self._end_date, _DEGRADATION_DATE):
            logger.warning(
                "USGS has reported degradation of eMODIS NDVI data "
                "due to issues with the MODIS sensor's satellite. "
                "This affects NDVI data from May 2022 (dekad 13), "
                "and users should be very careful of using any "
                "results after this date. This module is maintained "
                "for users to have access to historic data."
            )

        # warn if dates outside earliest dates
        if compare_dekads_lt(self._start_date, _EARLIEST_DATE):
            logger.warning(
                "Start date is before earliest date data is available. "
                f"Data will be downloaded from {_EARLIEST_DATE[0]}, dekad "
                f"{_EARLIEST_DATE[1]}."
            )

    def download(self, clobber: bool = False) -> Path:
        """Download raw NDVI data as .tif files.

        NDVI data is downloaded from the USGS API,
        with data for individual regions, years, and
        dekads stored as separate .tif files. No
        authentication is required. Data is downloaded
        for all available dekads from ``self._start_date``
        to ``self._end_date``.

        Parameters
        ----------
        clobber : bool, default = False
            If True, overwrites existing files

        Returns
        -------
        Path
            The downloaded filepath

        Examples
        --------
        >>> from aatoolbox import create_country_config, \
        ...  CodAB, UsgsNdviSmoothed
        >>>
        >>> # Retrieve admin 2 boundaries for Burkina Faso
        >>> country_config = create_country_config(iso3="bfa")
        >>>
        >>> # setup NDVI
        >>> bfa_ndvi = UsgsNdviSmoothed(
        ...     country_config=country_config,
        ...     start_date=[2020, 1],
        ...     end_date=[2020, 3]
        ... )
        >>> bfa_ndvi.download()
        """
        download_dekads = expand_dekads(
            dekad1=self._start_date, dekad2=self._end_date
        )
        for year, dekad in download_dekads:
            self._download_ndvi_dekad(year=year, dekad=dekad, clobber=clobber)
        return self._raw_base_dir

    def process(  # type: ignore
        self,
        gdf: gpd.GeoDataFrame,
        feature_col: str,
        clobber: bool = False,
        **kwargs,
    ) -> Path:
        """
        Process NDVI data for specific area.

        NDVI data is clipped to the provided
        ``geometries``, usually a geopandas
        dataframes ``geometry`` feature. ``kwargs``
        are passed on to ``aat.computer_raster_stats()``.
        The ``feature_col`` is used to define
        the unique processed file.

        The processing keeps track of the latest timestamp of
        when the raw raster files were modified. If the latest
        timestamp of the raw data is greater than when it was
        last processed, the file will automatically be
        re-processed.

        Parameters
        ----------
        gdf : geopandas.GeoDataFrame
            GeoDataFrame with row per area for stats computation.
            If ``pd.DataFrame`` is passed, geometry column must
            have the name ``geometry``. Passed to
            ``aat.compute_raster_stats()``.
        feature_col : str
            Column in ``gdf`` to use as row/feature identifier.
            and dates. Passed to ``aat.compute_raster_stats()``.
            The string is also used as a suffix to the
            processed file path for unique identication of
            analyses done on different files and columns.
        clobber : bool, default = False
            If True, overwrites existing processed dates. If
            the new file matches the old file, dates will be
            reprocessed and appended to the data frame. If
            files do not match, the old file will be replaced.
            If False, stats are only calculated for year-dekads
            that have not already been calculated within the
            file. However, if False and files do not match,
            value error will be raised.
        **kwargs
            Additional keyword arguments passed to
            ``aat.computer_raster_stats()``.

        Returns
        -------
        Path
            The processed path

        Examples
        --------
        >>> from aatoolbox import create_country_config, \
        ...  CodAB, UsgsNdviSmoothed
        >>>
        >>> # Retrieve admin 2 boundaries for Burkina Faso
        >>> country_config = create_country_config(iso3="bfa")
        >>> codab = CodAB(country_config=country_config)
        >>> bfa_admin2 = codab.load(admin_level=2)
        >>> bfa_admin1 = codab.load(admin_level=1)
        >>>
        >>> # setup NDVI
        >>> bfa_ndvi = UsgsNdviSmoothed(
        ...     country_config=country_config,
        ...     start_date=[2020, 1],
        ...     end_date=[2020, 3]
        ... )
        >>> bfa_ndvi.download()
        >>> bfa_ndvi.process(
        ...     gdf=bfa_admin2,
        ...     feature_col="ADM2_FR"
        ... )
        >>>
        >>> # process for admin1
        >>> bfa_ndvi.process(
        ...     gdf=bfa_admin1,
        ...     feature_col="ADM1_FR"
        ... )
        """
        # get statistics for processing
        stats_list = kwargs.pop(
            "stats_list", ["mean", "std", "min", "max", "sum", "count"]
        )
        percentile_list = kwargs.pop("percentile_list", None)

        if not stats_list and not percentile_list:
            raise ValueError(
                "At least one of `stats_list` or "
                "`percentile_list` must not be `None` "
                "when passed to `process()`."
            )

        # getting stats and percentiles for processing
        stats_list = [] if not stats_list else stats_list
        percentile_list = (
            []
            if not percentile_list
            else [str(percent) for percent in percentile_list]
        )
        process_stats = stats_list + percentile_list
        percentile_identifier = [False] * len(stats_list) + [True] * len(
            percentile_list
        )

        # get dates for processing
        all_dates_to_process = expand_dekads(
            dekad1=self._start_date, dekad2=self._end_date
        )

        for stat, is_percentile in zip(process_stats, percentile_identifier):
            self._process(
                clobber=clobber,
                gdf=gdf,
                feature_col=feature_col,
                dates_to_process=all_dates_to_process,
                stat=stat,
                is_percentile=is_percentile,
                kwargs=kwargs,
            )

        return self._processed_base_dir

    def load(self, feature_col: str) -> pd.DataFrame:  # type: ignore
        """
        Load the processed USGS NDVI data.

        Parameters
        ----------
        feature_col : str
            String is  used as a suffix to the
            processed file path for unique identication of
            analyses done on different files and columns.
            The same value must be passed to ``process()``.

        Returns
        -------
        pd.DataFrame
            The processed NDVI dataset.

        Raises
        ------
        FileNotFoundError
            If the requested file cannot be found.

        Examples
        --------
        >>> from aatoolbox import create_country_config, \
        ...  CodAB, UsgsNdviSmoothed
        >>>
        >>> # Retrieve admin 2 boundaries for Burkina Faso
        >>> country_config = create_country_config(iso3="bfa")
        >>> codab = CodAB(country_config=country_config)
        >>> bfa_admin2 = codab.load(admin_level=2)
        >>>
        >>> # setup NDVI
        >>> bfa_ndvi = UsgsNdviSmoothed(
        ...     country_config=country_config,
        ...     start_date=[2020, 1],
        ...     end_date=[2020, 3]
        ... )
        >>> bfa_ndvi.download()
        >>> bfa_ndvi.process(
        ...    gdf=bfa_admin2,
        ...    feature_col="ADM2_FR"
        )
        >>> bfa_ndvi.load(feature_col="ADM2_FR")
        """
        try:
            processed_files = self._find_processed_files(
                feature_col=feature_col
            )
            processed_dfs = [
                self._load(filepath=fp, drop_modified=True)
                for fp in processed_files
            ]

            df = functools.reduce(
                lambda df1, df2: pd.merge(
                    df1, df2, on=["date", "year", "dekad", feature_col]
                ),
                processed_dfs,
            )
        except TypeError as err:
            raise FileNotFoundError(
                "Files not found to load. Ensure the download() and process() "
                "methods are called prior to load()."
            ) from err

        # filter loaded data frame between our instances dates
        load_dates = expand_dekads(
            dekad1=self._start_date, dekad2=self._end_date
        )

        loaded_dates = df[["year", "dekad"]].values.tolist()
        keep_rows = [tuple(d) in load_dates for d in loaded_dates]

        df = df.loc[keep_rows]

        # put feature column as 1st column
        df_feature_col = df.pop(feature_col)
        df.insert(loc=0, column=feature_col, value=df_feature_col)

        return df

    def load_raster(
        self, load_date: Union[date, str, Tuple[int, int]]
    ) -> xr.DataArray:
        """Load raster for specific year and dekad.

        Parameters
        ----------
        load_date : Union[date, str, Tuple[int, int]]
            Date. Can be passed as a ``datetime.date``
            object and the relevant dekad will be determined,
            as a date string in ISO8601 format, or as a
            year-dekad tuple, i.e. (2020, 1).

        Returns
        -------
        xr.DataArray
            Data array of NDVI data.

        Raises
        ------
        FileNotFoundError
            If the requested file cannot be found.
        """
        year, dekad = get_dekadal_date(input_date=load_date)

        filepath = self._get_raw_path(year=year, dekad=dekad, local=True)
        try:
            da = rioxarray.open_rasterio(filepath)
            # get time file was updated
            file_time = datetime.fromtimestamp(filepath.stat().st_mtime)
            # assign coordinates for year/dekad
            # time dimension
            da = (
                da.assign_coords(
                    {
                        "year": year,
                        "dekad": dekad,
                        "date": dekad_to_date(dekad=(year, dekad)),
                        "modified": file_time,
                    }
                )
                .expand_dims("date")
                .squeeze("band", drop=True)
            )

            return da

        except RasterioIOError as err:
            # check if the requested date is outside the instance bounds
            # don't prevent loading, but use for meaningful error
            gt_end = compare_dekads_gt(
                dekad1=(year, dekad), dekad2=self._end_date
            )
            lt_start = compare_dekads_lt(
                dekad1=(year, dekad), dekad2=self._start_date
            )
            if gt_end or lt_start:
                file_warning = (
                    f"The requested year and dekad, {year}-{dekad}"
                    f"are {'greater' if gt_end else 'less'} than the "
                    f"instance {'end' if gt_end else 'start'} year and dekad, "
                    f"{self._end_date[0] if gt_end else self._start_date[0]}-"
                    f"{self._end_date[1] if gt_end else self._start_date[1]}. "
                    "Calling the `download()` method will not download this "
                    "file, and you need to re-instantiate the class to "
                    "include these dates."
                )
            else:
                file_warning = (
                    "Make sure that you have called the `download()` "
                    f"method and that the file {filepath.name} exists "
                    f"in {filepath.parent}."
                )
            raise FileNotFoundError(
                f"Cannot open the .tif file {filepath}. {file_warning}"
            ) from err

    def _download_ndvi_dekad(
        self, year: int, dekad: int, clobber: bool
    ) -> None:
        """Download NDVI for specific dekad.

        Parameters
        ----------
        year : int
            Year
        dekad : int
            Dekad
        clobber : bool
            If True, overwrites existing file
        """
        filepath = self._get_raw_path(year=year, dekad=dekad, local=True)
        url_filename = self._get_raw_filename(
            year=year, dekad=dekad, local=False
        )
        self._download(
            filepath=filepath, url_filename=url_filename, clobber=clobber
        )

    def _download(self, filepath: Path, url_filename: str, clobber: bool):
        local_filename = filepath.stem

        url = self._get_url(filename=url_filename)
        year, dekad = self._fp_year_dekad(filepath)

        try:
            resp = urlopen(url)
        except HTTPError:
            logger.error(
                f"No NDVI data available for "
                f"dekad {dekad} of {year}, skipping."
            )
            return

        if filepath.exists():
            file_time = datetime.fromtimestamp(filepath.stat().st_mtime)
            url_time_str = resp.headers["last-modified"]
            url_time = datetime.strptime(
                url_time_str,
                # Fri, 27 Mar 2015 08:05:42 GMT
                "%a, %d %b %Y %X %Z",
            )
            if not clobber and url_time <= file_time:
                logger.info(
                    f"File {filepath} exists, clobber set to False, and "
                    "file not been modified since last download. "
                    "Using existing files."
                )
                return filepath

        logger.info(
            f"Downloading NDVI data for {year}, dekad {dekad} "
            f"into {filepath}."
        )

        # open file within memory
        zf = ZipFile(BytesIO(resp.read()))

        # extract single .tif file from .zip
        for file in zf.infolist():
            if file.filename.endswith(".tif"):
                # rename the file to standardize to name of zip
                file.filename = f"{local_filename}.tif"
                zf.extract(file, self._raw_base_dir)

        resp.close()
        return filepath

    def _process(
        self,
        clobber: bool,
        gdf: gpd.GeoDataFrame,
        feature_col: str,
        dates_to_process: list,
        stat: str,
        is_percentile: bool,
        kwargs,
    ) -> Path:
        """Process data for particular statistic."""
        # get processed path for particular statistic
        percentile_list: Optional[List[int]]
        if is_percentile:
            stats_list = None
            percentile_list = [int(stat)]
            stat = f"{stat}quant"
        else:
            stats_list = [stat]
            percentile_list = None

        processed_path = self._get_processed_path(
            feature_col=feature_col, stat=stat
        )

        if processed_path.is_file():
            logger.info(
                f"Processing data from {self._start_date[0]}, "
                f"dekad {self._start_date[1]} to {self._end_date[0]} "
                f"dekad {self._end_date[1]} into {processed_path}."
            )

            (
                dates_to_process,
                df_already_processed,
            ) = self._determine_process_dates(
                clobber=clobber,
                filepath=processed_path,
                dates_to_process=dates_to_process,
            )

            if not dates_to_process:
                logger.info(
                    (
                        "No new {stat} data to process between "
                        f"{self._start_date[0]}, "
                        f"dekad {self._start_date[1]} "
                        f"and {self._end_date[0]}, "
                        f"dekad {self._end_date[1]}, "
                        "set `clobber = True` to re-process this data."
                    )
                )
                return processed_path
        else:
            # no dates already processed
            df_already_processed = pd.DataFrame()

        # process data for necessary dates
        data = [df_already_processed]
        for process_date in dates_to_process:
            da = self.load_raster(process_date)
            stats = da.aat.compute_raster_stats(
                gdf=gdf,
                feature_col=feature_col,
                stats_list=stats_list,
                percentile_list=percentile_list,
                **kwargs,
            )
            data.append(stats)

        # join data together and sort
        df = pd.concat(data).sort_values(by="date").reset_index(drop=True)

        # saving file
        self._processed_base_dir.mkdir(parents=True, exist_ok=True)
        df.to_csv(processed_path, index=False)

        return processed_path

    def _determine_process_dates(
        self, clobber: bool, filepath: Path, dates_to_process: list
    ) -> Tuple[list, pd.DataFrame]:
        """Determine dates to process.

        Parameters
        ----------
        clobber : bool
            If True, overwrites existing file
        filepath : Path
            Filepath to the existing processed file.
        dates_to_process : list
            List of dates to process

        Returns
        -------
        Tuple[list, pd.DataFrame]
            Returns a list of dates to process, filtered
            based on clobber, and a data frame of existing
            data to build upon in processing

        Raises
        ------
        ValueError
            Raised if not `clobber` but the statistics
            and `feature_col` for processing do not
            match the existing file.
        """
        df = self._load(filepath=filepath)

        # get dates that have already been processed
        dates_already_processed = df[
            ["year", "dekad", "modified"]
        ].values.tolist()
        dates_already_processed = {
            tuple(d[0:2]): d[2] for d in dates_already_processed
        }

        if clobber:
            # remove processed dates from file
            # so they can be reprocessed
            keep_rows = [
                d not in dates_to_process
                for d in list(dates_already_processed.keys())
            ]
            df = df.loc[keep_rows]
        else:
            # remove processed dates from dates to process
            dates_to_process = [
                d
                for d in dates_to_process
                if d not in list(dates_already_processed.keys())
                or self._get_modified_time(year=d[0], dekad=d[1])
                > dates_already_processed[d]
            ]

            # incase dates still processed based on
            # modified time remove from the df
            keep_rows = [
                d not in dates_to_process
                for d in list(dates_already_processed.keys())
            ]
            df = df.loc[keep_rows]
        return (dates_to_process, df)

    def _get_raw_filename(self, year: int, dekad: int, local: bool) -> str:
        """Get raw filename (excluding file type suffix).

        Parameters
        ----------
        year : int
            4-digit year
        dekad : int
            Dekad
        local : bool
            If True, returns filepath for local storage,
            which includes full 4-digit year and _
            separating with dekad. If False, filepath
            corresponds to the zip file stored in the
            USGS server.

        Returns
        -------
        str
            File path prefix for .zip file at URL and
            for .tif files stored within the .zip
        """
        if local:
            file_year = f"{year:04}_"
        else:
            file_year = f"{year-2000:02}"
        file_name = (
            f"{self._datasource_config.area_prefix}{file_year}"
            f"{dekad:02}{self._data_variable_suffix}"
        )
        return file_name

    @staticmethod
    def _load(filepath: Path, drop_modified: bool = False):
        """Load processed data.

        Parameters
        ----------
        filepath : Path
            Filepath to processed data
        drop_modified : bool, default = False
            Drop modified column, used for merging data frames
            together.

        Returns
        -------
        pd.DataFrame
            Processed data frame
        """
        df = pd.read_csv(filepath, parse_dates=["date", "modified"])
        if drop_modified:
            df.drop(["modified"], axis=1, inplace=True)
        return df

    def _get_raw_path(self, year: int, dekad: int, local: bool) -> Path:
        """Get raw filepath.

        Parameters
        ----------
        year : int
            4-digit year
        dekad : int
            Dekad
        local : bool
            If True, returns filepath for local storage,
            which includes full 4-digit year and _
            separating with dekad. If False, filepath
            corresponds to the zip file stored in the
            USGS server.

        Returns
        -------
        Path
            Path to raw file
        """
        filename = self._get_raw_filename(year=year, dekad=dekad, local=local)
        return self._raw_base_dir / f"{filename}.tif"

    def _get_modified_time(self, year: int, dekad: int) -> datetime:
        """Get modified time of raw file.

        Used to determine when to re-process
        or re-download raw raster files.

        Parameters
        ----------
        year : int
            4-digit year
        dekad : int
            Dekad

        Returns
        -------
        datetime
            Timestamp of when file was modified.
        """
        filepath = self._get_raw_path(year=year, dekad=dekad, local=True)
        return datetime.fromtimestamp(filepath.stat().st_mtime)

    def _get_processed_filename(self, feature_col: str, stat: str) -> str:
        """Return processed filename.

        Returns the processed filename. The suffix
        of the filename is always the ``feature_col``
        the statistics are aggregated to.

        Returns
        -------
        str
            Processed filename
        """
        base_file_name = self._get_processed_base_filename(
            feature_col=feature_col
        )
        return f"{base_file_name}_{stat}.csv"

    def _get_processed_base_filename(self, feature_col: str) -> str:
        base_file_name = (
            f"{self._country_config.iso3}"
            f"_usgs_ndvi_{self._data_variable}"
            f"_{feature_col}"
        )
        return base_file_name

    def _get_processed_path(self, feature_col: str, stat: str) -> Path:
        return self._processed_base_dir / self._get_processed_filename(
            feature_col=feature_col, stat=stat
        )

    def _find_processed_files(self, feature_col: str) -> Generator:
        base_file_name = self._get_processed_base_filename(
            feature_col=feature_col
        )
        return self._processed_base_dir.glob(f"{base_file_name}_*.csv")

    def _get_url(self, filename) -> str:
        """Get USGS NDVI URL.

        Parameters
        ----------
        filename : str
            File name string generated for specific year, dekad, and data type

        Returns
        -------
        str
            Download URL string
        """
        return (
            f"https://edcintl.cr.usgs.gov/downloads/sciweb1/shared/fews/web/"
            f"{self._datasource_config.area_url}/dekadal/emodis"
            f"/ndvi_c6/{self._data_variable_url}/"
            f"downloads/dekadal/{filename}.zip"
        )

    # TODO: potentially move from static method to
    # wider USGS function repository
    @staticmethod
    def _fp_year_dekad(path: Path) -> List[int]:
        """Extract year and dekad from filepath.

        Parameters
        ----------
        path : Path
            Filepath

        Returns
        -------
        list
            List of year and dekad
        """
        filename = path.stem
        # find two groups, first for year second for dekad
        regex = re.compile(r"(\d{4})_(\d{2})")
        return [int(x) for x in regex.findall(filename)[0]]



"""Test the GloFAS instantiation."""
from datetime import date

import pytest

from aatoolbox import GlofasForecast, GlofasReanalysis, GlofasReforecast


def test_reanalysis_dates(mock_country_config, geo_bounding_box):
    """Test date range behaviour of reanalysis."""

    def glofas_reanalysis(end_date: date, start_date: date = None):
        return GlofasReanalysis(
            country_config=mock_country_config,
            geo_bounding_box=geo_bounding_box,
            start_date=start_date,
            end_date=end_date,
        )

    # Try using a string as a date
    good_end_date = "2022-11-15"
    # These should not throw an error
    glofas_reanalysis(end_date=good_end_date)
    # End date too far in future
    glofas_future = glofas_reanalysis(end_date=date(year=3000, month=1, day=1))
    assert glofas_future._end_date == date.today()
    # Start date too early
    glofas_past = glofas_reanalysis(
        start_date=date(year=1800, month=1, day=1), end_date=good_end_date
    )
    assert glofas_past._start_date == date(year=1979, month=1, day=1)
    # End date > start date
    with pytest.raises(ValueError):
        glofas_reanalysis(
            start_date=date(year=2020, month=1, day=2),
            end_date=date(year=2020, month=1, day=1),
        )


def test_forecast_dates(mock_country_config, geo_bounding_box):
    """Test date range behaviour of forecast."""

    def glofas_forecast(end_date: date, start_date: date = None):
        return GlofasForecast(
            country_config=mock_country_config,
            geo_bounding_box=geo_bounding_box,
            start_date=start_date,
            end_date=end_date,
            leadtime_max=15,
        )

    # Try using a string as a date
    good_end_date = "2022-11-15"
    # These should not throw an error
    glofas_forecast(end_date=good_end_date)
    # End date too far in future
    glofas_future = glofas_forecast(end_date=date(year=3000, month=1, day=1))
    assert glofas_future._end_date == date.today()
    # Start date too early
    glofas_past = glofas_forecast(
        start_date=date(year=1800, month=1, day=1), end_date=good_end_date
    )
    assert glofas_past._start_date == date(year=2021, month=5, day=26)
    # End date > start date
    with pytest.raises(ValueError):
        glofas_forecast(
            start_date=date(year=2020, month=1, day=2),
            end_date=date(year=2020, month=1, day=1),
        )


def test_reforecast_dates(mock_country_config, geo_bounding_box):
    """Test date range behaviour of reforecast."""

    def glofas_reforecast(start_date: date = None, end_date: date = None):
        return GlofasReforecast(
            country_config=mock_country_config,
            geo_bounding_box=geo_bounding_box,
            start_date=start_date,
            end_date=end_date,
            leadtime_max=15,
        )

    # These should not throw an error
    glofas_reforecast()
    # Try using a string as a date
    glofas_reforecast(end_date="2010-01-01")
    # End date too far in future
    glofas_future = glofas_reforecast(end_date=date(year=3000, month=1, day=1))
    assert glofas_future._end_date == date(year=2018, month=12, day=31)
    # Start date too early
    glofas_past = glofas_reforecast(start_date=date(year=1800, month=1, day=1))
    assert glofas_past._start_date == date(year=1999, month=1, day=1)
    # End date > start date
    with pytest.raises(ValueError):
        glofas_reforecast(
            start_date=date(year=2020, month=1, day=2),
            end_date=date(year=2020, month=1, day=1),
        )



"""Function for checking file existence."""
import logging
from typing import Any, Callable, Optional, TypeVar

import wrapt

from aatoolbox.datasources.datasource import DataSource

logger = logging.getLogger(__name__)

# For typing the decorator
F = TypeVar("F", bound=Callable[..., Any])


@wrapt.decorator
def check_file_existence(
    wrapped: F, instance: Optional[DataSource], args: list, kwargs: dict
) -> F:
    """
    Don't overwrite existing data.

    Avoid recreating data if it already exists and if clobber not
    toggled by user. Used to wrap functions that accept filepath
    as a keyword argument.

    Parameters
    ----------
    wrapped : function
        The function to wrap. The function must have "filepath" as
        a keyword parameter, and it can also have an optional
        "clobber" boolean keyword parameter.
    instance : Optional[DataSource]
        Object the wrapped function is bound to. Not used within, but
        ensures that instance methods do not pass `self` to args.
    args : list
        List of positional arguments.
    kwargs : dict
        Dictionary of keyword arguments

    Returns
    -------
    If filepath exists and clobber is False, returns filepath.
    Otherwise, returns the result of the decorated function.

    Raises
    ------
    KeyError
        If `filepath` or `clobber` are not passed as kwargs.
    """
    try:
        filepath = kwargs["filepath"]
        clobber = kwargs["clobber"]
    except KeyError as err:
        raise KeyError(
            (
                "`filepath` and `clobber` must be passed as keyword "
                "arguments for the `check_file_existence`"
                " decorator to work."
            )
        ) from err

    # check existence
    exist_dict = {True: "exists", False: "does not exist"}

    # check filepath exists -> clobber
    usage_dict = {
        True: {True: "overwriting existing", False: "using existing"},
        False: {True: "downloading new", False: "downloading new"},
    }
    fp_exists = filepath.exists()

    logger.info(
        f"File {filepath} {exist_dict[fp_exists]} and clobber "
        f"set to {clobber}, {usage_dict[fp_exists][clobber]} file."
    )

    if fp_exists and not clobber:
        return filepath
    else:
        return wrapped(*args, **kwargs)



"""Base class for aatoolbox data source."""
from abc import ABC, abstractmethod
from pathlib import Path

from aatoolbox.config.countryconfig import CountryConfig
from aatoolbox.config.pathconfig import PathConfig

_GLOBAL_DIR = "glb"


class DataSource(ABC):
    """
    Base abstract class object that contains path convenience functions.

    Cannot itself be instantiated. ``__init__``, ``download()``,
    ``load()``, and ``process()`` methods required for subclass to be
    instantiated.

    Parameters
    ----------
    country_config: CountryConfig
        Country configuration
    datasource_base_dir : str
        Module directory name (usually correspond to data source)
    is_public: bool, default = False
        Whether the dataset is public or private. Determines top-level
        directory structure.
    is_global_raw: bool, default = False
        Whether the raw dataset should be saved in the `glb` folder. This is
        normally done when it has global or regional coverage.
    is_global_processed: bool, default = False
        Whether the processed dataset should be saved in the `glb` folder.
        This is normally done when it has global or regional coverage.
    config_datasource_name: str = None
        The name of the attribute in the config file
    """

    @abstractmethod
    def __init__(
        self,
        country_config: CountryConfig,
        datasource_base_dir: str,
        is_public: bool = False,
        is_global_raw: bool = False,
        is_global_processed: bool = False,
        config_datasource_name: str = None,
    ):
        if config_datasource_name is not None:
            self._datasource_config = self._config_attribute_name_validator(
                config_datasource_name=config_datasource_name,
                country_config=country_config,
            )
        self._country_config = country_config
        self._datasource_base_dir = datasource_base_dir
        self._path_config = PathConfig()
        self._raw_base_dir = self._get_base_dir(
            is_public=is_public, is_raw=True, is_global=is_global_raw
        )
        self._processed_base_dir = self._get_base_dir(
            is_public=is_public, is_raw=False, is_global=is_global_processed
        )

    @staticmethod
    def _config_attribute_name_validator(
        config_datasource_name: str, country_config: CountryConfig
    ):
        try:
            datasource_config = getattr(country_config, config_datasource_name)
        except AttributeError:
            datasource_config = None
        # If the datasource is one of the defaults, it's set to None and
        # thus an attribute error won't be raised. So also need to check for
        # the case when it is None.
        if datasource_config is None:
            raise AttributeError(
                f"{config_datasource_name} needs to be added to the "
                f"config file. See the documentation for more details."
            )
        return datasource_config

    def _get_base_dir(
        self,
        is_public: bool,
        is_raw: bool,
        is_global: bool = False,
    ) -> Path:
        """
        Define the base_dir.

        Parameters
        ----------
        is_public: bool
            Whether the dataset is public or private. Determines top-level
            directory structure.
        is_raw: bool
            Whether the dataset is raw or processed
        is_global: bool
            Whether the dataset is global (or regional) or specific to the
            iso3
        """
        permission_dir = (
            self._path_config.public
            if is_public
            else self._path_config.private
        )
        state_dir = (
            self._path_config.raw if is_raw else self._path_config.processed
        )
        region_dir = (
            self._country_config.iso3 if not is_global else _GLOBAL_DIR
        )
        return (
            self._path_config.base_path
            / permission_dir
            / state_dir
            / region_dir
            / self._datasource_base_dir
        )

    @abstractmethod
    def download(self, clobber: bool = False):
        """Abstract method for downloading."""
        pass

    @abstractmethod
    def process(self, clobber: bool = False):
        """Abstract method for processing."""
        pass

    @abstractmethod
    def load(self):
        """Abstract method for loading."""
        pass



"""Tests for HDX API utility."""
from collections import UserDict

import pytest

from aatoolbox.utils.hdx_api import load_dataset_from_hdx


@pytest.fixture(autouse=True)
def mock_resource(mocker):
    """Mock the HDX download function."""
    # Resource is a UserDict so need to make a class to mock that
    class MockResource(UserDict):
        def download(self, folder):
            return "", "resource_filepath"

    mock_dataset = mocker.patch(
        "aatoolbox.utils.hdx_api.Dataset.read_from_hdx"
    )
    # read_from_hdx creates an instance, need to mock the instance
    # method get_resources to return custom resources
    mock_dataset.return_value.get_resources.return_value = [
        MockResource({"name": "resource1"})
    ]
    # Also need to mock out shutil
    mocker.patch("aatoolbox.utils.hdx_api.shutil")


def test_returns_filepath(tmp_path):
    """Test that querying HDX API returns expected filepath."""
    input_filepath = tmp_path / "hdx_test_path"
    output_filepath = load_dataset_from_hdx(
        hdx_address="hdx_address",
        hdx_dataset_name="resource1",
        output_filepath=input_filepath,
    )
    assert output_filepath == input_filepath


def test_error_when_not_found(tmp_path):
    """Test that missing resource raises error."""
    with pytest.raises(FileNotFoundError):
        load_dataset_from_hdx(
            hdx_address="hdx_address",
            hdx_dataset_name="some_name_not_in_fake_resrouce",
            output_filepath=tmp_path / "hdx_test_error",
        )



"""Base class for downloading and processing GloFAS river discharge data."""
import logging
import time
from abc import abstractmethod
from dataclasses import dataclass
from datetime import date
from pathlib import Path
from typing import List, Tuple, Union

import cdsapi
import numpy as np
import xarray as xr
from dateutil import rrule

from aatoolbox.config.countryconfig import CountryConfig
from aatoolbox.datasources.datasource import DataSource
from aatoolbox.utils.dates import get_date_from_user_input
from aatoolbox.utils.geoboundingbox import GeoBoundingBox

_MODULE_BASENAME = "glofas"
_HYDROLOGICAL_MODEL = "lisflood"
_RIVER_DISCHARGE_VAR = "dis24"
_REQUEST_SLEEP_TIME = 60  # seconds

logger = logging.getLogger(__name__)


@dataclass
class _QueryParams:
    """
    Class to keep track of CDS query input and output.

    Parameters
    ----------
    filepath: Path
        Full filepath of downloaded CDS file
    query: dict
        Output of _get_query() method, to be submitted to the CDS API
    request_id: str, default = None
        Request ID from CDS, only set after request has been made
    downloaded: bool, default = False
        Whether or not the file has yet been downloaded
    """

    filepath: Path
    query: dict
    request_id: str = None  # type: ignore # TODO: fix
    downloaded: bool = False


class Glofas(DataSource):
    """
    Base class for all GloFAS data downloading and processing.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    geo_bounding_box: GeoBoundingBox
        The bounding coordinates of the area that should be included
    cds_name : str
        The name of the dataset in CDS
    system_version : str
        An input to the CDS query
    product_type : str or list
        Which product types from the dataset are requested
    date_variable_prefix : str
        Some dates require a prefix for the CDS API query
    frequency : str
        How to split the query (and thus files): in years, months, or days.
        Depends on the maximum query size of the product
    coord_names : list
        Coordinate names in the xarray dataset
    start_date_min: date
        The minimum allowed start date
    end_date_max: date, default = None
        The maximum allowed end date
    start_date : Union[date, str], default = None
        The starting date for the dataset
    end_date : Union[date, str], default = None
        The ending date for the dataset
    leadtime_max : int, default = None
        The maximum lead time in days, for forecast or reforecast data
    """

    def __init__(
        self,
        country_config: CountryConfig,
        geo_bounding_box: GeoBoundingBox,
        cds_name: str,
        system_version: str,
        product_type: Union[str, List[str]],
        date_variable_prefix: str,
        frequency: int,
        coord_names: List[str],
        start_date_min: date,
        end_date_max: date = None,
        start_date: Union[date, str] = None,
        end_date: Union[date, str] = None,
        leadtime_max: int = None,
    ):
        super().__init__(
            country_config=country_config,
            datasource_base_dir=_MODULE_BASENAME,
            is_public=True,
        )
        # The GloFAS API on CDS requires coordinates have the format x.x5
        self._geo_bounding_box = geo_bounding_box.round_coords(
            offset_val=0.05, round_val=0.1
        )
        self._start_date, self._end_date = _set_dates(
            start_date_min=start_date_min,
            end_date_max=end_date_max,
            start_date=start_date,
            end_date=end_date,
        )
        self._cds_name = cds_name
        self._system_version = system_version
        self._product_type = product_type
        self._date_variable_prefix = date_variable_prefix
        self._frequency = frequency
        self._coord_names = coord_names
        self._leadtime_max = leadtime_max
        self._forecast_type = type(self).__name__.replace("Glofas", "").lower()
        self._date_range = rrule.rrule(
            freq=self._frequency,
            dtstart=self._start_date,
            until=self._end_date,
        )

    def download(  # type: ignore
        self,
        clobber: bool = False,
    ) -> List[Path]:
        """
        Download the GloFAS data by querying CDS.

        The raw GloFAS data is available as a global raster in CDS. This method
        downloads the raster files for the specified region of interest
        and date range. The files are in GRIB format and are split up either
        by day, month, or year depending on the GloFAS product.

        Parameters
        ----------
        clobber : bool, default = False
            Overwrite files that were already downloaded

        Returns
        -------
        A list paths of downloaded files

        """
        msg = (
            f"Downloading GloFAS {self._forecast_type} "
            f"for {self._start_date} - {self._end_date}"
        )
        if self._leadtime_max is not None:
            msg += f"and up to {self._leadtime_max} day lead time"
        logger.info(msg)

        # Make directory
        output_directory = self._get_directory()
        output_directory.mkdir(parents=True, exist_ok=True)

        # Get list of files to open
        query_params_list = []
        for file_date in self._date_range:
            output_filepath = self._get_filepath(
                year=file_date.year,
                month=file_date.month,
                day=file_date.day,
            )
            if not clobber and output_filepath.exists():
                continue
            query_params_list.append(
                _QueryParams(
                    filepath=output_filepath,
                    query=self._get_query(
                        year=file_date.year,
                        month=file_date.month,
                        day=file_date.day,
                    ),
                )
            )

        download_filepaths = self._download(
            query_params_list=query_params_list
        )
        logger.info(
            f"Downloaded {len(download_filepaths)} files to {output_directory}"
        )
        logger.debug(f"Files downloaded: {download_filepaths}")
        return download_filepaths

    def process(  # type: ignore
        self,
        clobber: bool = False,
    ) -> List[Path]:
        """
        Process the downloaded GloFAS files.

        For each raw GRIB file, read it in and extract the river discharge
        from the reporting point coordinates specified in the configuration
        file. Saves the output as a NetCDF file, where files are split by day,
        month or year depending on the GloFAS product.

        Parameters
        ----------
        clobber : bool, default = False
            Overwrite files that were already processed

        Returns
        -------
        A list paths of processed files

        """
        logger.info(
            f"Processing GloFAS {self._forecast_type} for "
            f"{self._start_date} - {self._end_date} and up to "
            f"{self._leadtime_max} day lead time"
        )
        # Make the directory
        output_directory = self._get_directory(is_processed=True)
        output_directory.mkdir(parents=True, exist_ok=True)
        # Get list of files to open
        processed_filepaths = []
        for file_date in self._date_range:
            input_filepath = self._get_filepath(
                year=file_date.year,
                month=file_date.month,
                day=file_date.day,
            )
            output_filepath = self._get_filepath(
                year=file_date.year,
                month=file_date.month,
                day=file_date.day,
                is_processed=True,
            )
            if not clobber and output_filepath.exists():
                continue
            logger.debug(f"Processing {input_filepath}")
            ds_raw = self._load_single_file(
                input_filepath=input_filepath,
                filepath=output_filepath,
                clobber=clobber,
            )
            ds_processed = self._get_reporting_point_dataset(ds=ds_raw)
            # NetCDF doesn't like to overwrite files
            if output_filepath.exists():
                output_filepath.unlink()
            ds_processed.to_netcdf(output_filepath)
            processed_filepaths.append(output_filepath)
            logger.debug(f"Wrote file to {output_filepath}")
        logger.info(
            f"Processed {len(processed_filepaths)} files to {output_directory}"
        )
        logger.debug(f"Files downloaded: {processed_filepaths}")
        return processed_filepaths

    def load(
        self,
    ) -> xr.Dataset:
        """
        Load the processed GloFAS data as an xarray.DataSet.

        Returns
        -------
        A single xarray dataset containing all GloFAS reporting points
        and their associated river discharge

        """
        filepath_list = [
            self._get_filepath(
                year=dataset_date.year,
                month=dataset_date.month,
                day=dataset_date.day,
                is_processed=True,
            )
            for dataset_date in self._date_range
        ]
        with xr.open_mfdataset(
            filepath_list, preprocess=self._preprocess_load
        ) as ds:
            return ds

    def _get_filepath(
        self,
        year: int,
        month: int = None,
        day: int = None,
        is_processed: bool = False,
    ) -> Path:
        """Get downloaded / processed filepaths based on GloFAS product."""
        filename = f"{self._country_config.iso3}_{self._cds_name}_{year}"
        if self._frequency in [rrule.MONTHLY, rrule.DAILY]:
            filename += f"-{str(month).zfill(2)}"
        if self._frequency == rrule.DAILY:
            filename += f"-{str(day).zfill(2)}"
        if self._leadtime_max is not None:
            filename += f"_ltmax{str(self._leadtime_max).zfill(2)}d"
        filename += f"_{self._geo_bounding_box.get_filename_repr(p=2)}"
        if is_processed:
            filename += "_processed.nc"
        else:
            filename += ".grib"
        return self._get_directory(is_processed=is_processed) / Path(filename)

    def _get_directory(self, is_processed: bool = False) -> Path:
        """Get download / processed directory for GloFAS product."""
        return (
            self._processed_base_dir
            if is_processed
            else self._raw_base_dir / self._cds_name
        )

    def _download(
        self,
        query_params_list: List[_QueryParams],
    ) -> List[Path]:
        """
        Download the GloFAS data from CDS.

        Uses query_params_list, which is a list of API request input dicts,
        to query the CDS API, and for each query,  store the request
        ID that is returned, and the downloaded state.

        Then loops through the list of request, checking each one to see
        if it has been completed on the CDS side. If so, it's downloaded,
        and then removed from the list of requests. The process continues
        until the request list is empty
        """
        # First make the requests to the CDS client and store request number
        for query_params in query_params_list:
            logger.debug(f"Making request {query_params.query}")
            query_params.request_id = (
                cdsapi.Client(wait_until_complete=False, delete=False)
                .retrieve(name=self._cds_name, request=query_params.query)
                .reply["request_id"]
            )
        # Loop through the request list and check status until all requests
        # are downloaded
        downloaded_filepaths = []
        while query_params_list:
            for query_params in query_params_list:
                result = cdsapi.api.Result(
                    client=cdsapi.Client(
                        wait_until_complete=False, delete=False
                    ),
                    reply={"request_id": query_params.request_id},
                )
                result.update()
                state = result.reply["state"]
                logger.debug(
                    f"For request {query_params.request_id} and filename "
                    f"{query_params.filepath}, state is {state}"
                )
                if state == "completed":
                    result.download(query_params.filepath)
                    query_params.downloaded = True
                    downloaded_filepaths.append(query_params.filepath)
                elif state == "failed":
                    raise RuntimeError("Query has failed, try again")
            # Remove requests that have been downloaded
            query_params_list = [
                query_params
                for query_params in query_params_list
                if not query_params.downloaded
            ]
            # Sleep a bit before the next loop so that we're not
            # hammering on cds
            if query_params_list:
                time.sleep(_REQUEST_SLEEP_TIME)
        return downloaded_filepaths

    def _get_query(
        self,
        year: int,
        month: int = None,
        day: int = None,
    ) -> dict:
        """Create dictionary for CDS API query input."""
        query = {
            "variable": "river_discharge_in_the_last_24_hours",
            "format": "grib",
            "product_type": self._product_type,
            "system_version": self._system_version,
            "hydrological_model": _HYDROLOGICAL_MODEL,
            f"{self._date_variable_prefix}year": str(year),
            f"{self._date_variable_prefix}month": str(month).zfill(2)
            if self._frequency in [rrule.MONTHLY, rrule.DAILY]
            else [str(x + 1).zfill(2) for x in range(12)],
            f"{self._date_variable_prefix}day": str(day).zfill(2)
            if self._frequency == rrule.DAILY
            else [str(x + 1).zfill(2) for x in range(31)],
            "area": [
                self._geo_bounding_box.lat_max,
                self._geo_bounding_box.lon_min,
                self._geo_bounding_box.lat_min,
                self._geo_bounding_box.lon_max,
            ],
        }
        if self._leadtime_max is not None:
            leadtime = list(np.arange(self._leadtime_max) + 1)
            query["leadtime_hour"] = [
                str(single_leadtime * 24) for single_leadtime in leadtime
            ]
        logger.debug(f"Query: {query}")
        return query

    @abstractmethod
    def _load_single_file(self, *args, **kwargs) -> xr.Dataset:
        """Process a single raw raster file."""
        pass

    def _get_reporting_point_dataset(self, ds: xr.Dataset) -> xr.Dataset:
        """Convert raw raster to processed that uses reporting points."""
        if self._country_config.glofas is None:
            raise KeyError(
                "The country configuration file does not contain "
                "any reporting point coordinates. Please update the "
                "configuration file and try again."
            )
        # Check that lat and lon of reporting points are in the bounds
        for reporting_point in self._country_config.glofas.reporting_points:
            if (
                not ds.longitude.min()
                < reporting_point.lon
                < ds.longitude.max()
            ):
                raise IndexError(
                    f"ReportingPoint {reporting_point.id} has out-of-bounds "
                    f"lon value of {reporting_point.lon} (data lon ranges "
                    f"from {ds.longitude.min().values} "
                    f"to {ds.longitude.max().values})"
                )
            if not ds.latitude.min() < reporting_point.lat < ds.latitude.max():
                raise IndexError(
                    f"ReportingPoint {reporting_point.id} has out-of-bounds "
                    f"lat value of {reporting_point.lat} (data lat ranges "
                    f"from {ds.latitude.min().values} "
                    f"to {ds.latitude.max().values})"
                )
        # If reporting points fit then return processed dataset
        return xr.Dataset(
            data_vars={
                reporting_point.name: (
                    self._coord_names,
                    ds.sel(
                        longitude=reporting_point.lon,
                        latitude=reporting_point.lat,
                        method="nearest",
                    )[_RIVER_DISCHARGE_VAR].data,
                )
                # fmt: off
                for reporting_point in
                self._country_config.glofas.reporting_points
                # fmt: on
            },
            coords={
                coord_name: ds[coord_name] for coord_name in self._coord_names
            },
        )

    @staticmethod
    def _preprocess_load(ds: xr.Dataset) -> xr.Dataset:
        """Preprocessing to do before loading the processed data."""
        return ds


def _set_dates(
    start_date_min: date,
    end_date_max: date = None,
    start_date: Union[date, str] = None,
    end_date: Union[date, str] = None,
) -> Tuple[date, date]:
    """Adjust date types and check against limits."""
    # TODO: perhaps this is general enough to be useful for other data sources?
    # Set any missing dates
    if end_date_max is None:
        end_date_max = date.today()
    if start_date is None:
        start_date = start_date_min
    if end_date is None:
        end_date = end_date_max
    # Make sure that dates are all date type
    start_date = get_date_from_user_input(start_date)
    end_date = get_date_from_user_input(end_date)
    # Check against bounds
    if start_date > end_date:
        raise ValueError(
            "Please ensure that the start date is <= the end_date"
        )
    if start_date < start_date_min:
        logger.warning(
            f"Start date {start_date} is too far in the past,"
            f"setting to {start_date_min}"
        )
        start_date = start_date_min
    if end_date > end_date_max:
        logger.warning(
            f"End date {end_date} is too far in the future,"
            f"setting to {end_date_max}"
        )
        end_date = end_date_max
    return start_date, end_date



"""Function for I/O."""
import logging
import zipfile
from pathlib import Path
from typing import Union

import requests
import yaml

logger = logging.getLogger(__name__)


def download_url(
    url: str,
    save_path: Path,
    chunk_size: int = 2048,
):
    """
    Download the file located at `url` to `save_path`.

    Parameters
    ----------
    url : str
        url that contains the file to be downloaded
    save_path : Path
        path to the location the file should be saved
    chunk_size : int
        number of bytes to save at once
    """
    save_path.parent.mkdir(exist_ok=True, parents=True)
    # Remove file if already exists
    # TODO: Use missing_ok=True once 3.7 is dropped
    if save_path.exists():
        save_path.unlink()

    # use a session and chunk_size to prevent
    # crashing when downloading large files while
    # not loosing too much speed
    session = requests.Session()
    r = session.get(url, stream=True)
    r.raise_for_status()
    with save_path.open("wb") as fd:
        for chunk in r.iter_content(chunk_size=chunk_size):
            fd.write(chunk)


def unzip(
    zip_file_path: Path,
    save_dir: Path,
):
    """
    Unzip a file.

    Parameters
    ----------
    zip_file_path : Path
        path to the location the zip file is saved
    save_dir : Path
        dir path to which the content of the zip
        file should be saved
    """
    with zipfile.ZipFile(file=zip_file_path, mode="r") as zip_ref:
        zip_ref.extractall(save_dir)


def parse_yaml(filename: Union[str, Path]) -> dict:
    """
    Read in a yaml file.

    Parameters
    ----------
    filename : str, Path
    The full filepath of the YAML file

    Returns
    -------
    A dictionary with the YAML file contents
    """
    with open(file=filename, mode="r") as stream:
        config = yaml.safe_load(stream)
    return config



"""Tests for the raster utilities module."""
import doctest
import logging
import unittest

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from pandas._testing import assert_frame_equal
from rioxarray.exceptions import DimensionError, MissingCRS
from shapely.geometry import Polygon

import aatoolbox


def test_doctest_suite():
    """
    Test docstrings in raster module.

    Checks if there are any failures in the doctest
    suite, meaning that there is some error within
    the docstrings of the raster module. This is used
    to avoid running doctest on all docstrings in the
    library, which would require many exceptions given
    the extent of downloading and processing other modules
    rely on.

    Method found in:
    https://vladyslav-krylasov.medium.com/discover-unit-tests-and-doctests-in-one-run-c5504aea86bd
    """
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(aatoolbox.utils.raster))
    runner = unittest.TextTestRunner(verbosity=2).run(suite)
    assert not runner.failures


@pytest.fixture
def gdf():
    """Create geodataframe for testing."""
    d = {
        "name": ["area_a", "area_b"],
        "geometry": [
            Polygon([(0, 0), (0, 2), (2, 2), (2, 0)]),
            Polygon([(2, 0), (2, 2), (3, 2), (3, 0)]),
        ],
    }
    return gpd.GeoDataFrame(d)


@pytest.fixture
def gdf_missing(gdf):
    """Create geodataframe for testing with area not overlapping raster."""
    d = {
        "name": ["area_c"],
        "geometry": [Polygon([(3, 0), (3, 2), (4, 2), (4, 0)])],
    }
    df_missing = pd.concat([gdf, pd.DataFrame(d)], ignore_index=True)
    return gpd.GeoDataFrame(df_missing)


@pytest.fixture
def da_2d():
    """Create 2d raster array."""
    da = xr.DataArray(
        [[1, 2, 3], [4, 5, 6]],
        dims=("y", "x"),
        coords={"y": [1.5, 0.5], "x": [0.5, 1.5, 2.5]},
    ).rio.write_crs("EPSG:4326")
    return da


@pytest.fixture
def da_3d():
    """Create 3d raster array."""
    da = xr.DataArray(
        [
            [[1, 2, 3], [4, 5, 6]],
            [[1, 2, 3], [4, 5, 6]],
        ],
        dims=("time", "y", "x"),
        coords={
            "time": ["2020-01-01", "2020-01-02"],
            "y": [1.5, 0.5],
            "x": [0.5, 1.5, 2.5],
        },
    ).rio.write_crs("EPSG:4326")
    return da


@pytest.fixture
def ds_3d(da_3d):
    """Create 3d raster dataset."""
    return da_3d.to_dataset(name="val")


@pytest.fixture
def expected_2d():
    """Create expected comped stats for 2d raster."""
    df = pd.DataFrame(
        {
            "mean": {0: 3.0, 1: 4.5},
            "std": {0: 1.5811388300841898, 1: 1.5},
            "min": {0: 1, 1: 3},
            "max": {0: 5, 1: 6},
            "sum": {0: 12.0, 1: 9.0},
            "count": {0: 4, 1: 2},
            "name": {0: "area_a", 1: "area_b"},
        },
    )
    return df


@pytest.fixture
def expected_3d(expected_2d):
    """Create expected comped stats for 3d raster."""
    df_3d = pd.DataFrame(np.repeat(expected_2d.values, 2, axis=0))
    df_3d.columns = expected_2d.columns
    df_3d.insert(loc=0, column="time", value=["2020-01-01", "2020-01-02"] * 2)
    return df_3d


def test_compute_raster_stats_2d(da_2d, gdf, expected_2d):
    """Compute raster stats without time dimensions."""
    result = da_2d.aat.compute_raster_stats(gdf, "name")
    assert_frame_equal(result, expected_2d, check_dtype=False)
    expected_2d.insert(loc=6, column="95quant", value=[4.85, 5.85])
    result_pct = da_2d.aat.compute_raster_stats(
        gdf=gdf, feature_col="name", percentile_list=[95]
    )
    assert_frame_equal(result_pct, expected_2d, check_dtype=False)


def test_compute_raster_stats_3d(ds_3d, gdf, expected_3d):
    """Compute raster stats with time dimensions."""
    result = ds_3d.aat.compute_raster_stats(
        var_names=["val"], gdf=gdf, feature_col="name"
    )
    assert_frame_equal(result[0], expected_3d, check_dtype=False)
    result_all_vars = ds_3d.aat.compute_raster_stats(
        gdf=gdf, feature_col="name"
    )
    assert_frame_equal(result_all_vars[0], expected_3d, check_dtype=False)
    result_str = ds_3d.aat.compute_raster_stats(
        var_names="val", gdf=gdf, feature_col="name"
    )
    assert_frame_equal(result_str[0], expected_3d, check_dtype=False)


def test_compute_raster_stats_da_assertions(da_3d, gdf):
    """Ensure error assertions working in compute raster stats."""
    with pytest.raises(MissingCRS):
        da_3d.rio._crs = False
        da_3d.aat.compute_raster_stats(gdf=gdf, feature_col="name")


def test_set_time_dim(da_2d):
    """Ensure error assertions working in set time dim."""
    with pytest.raises(DimensionError):
        da_2d.aat.set_time_dim("time")


def test_correct_calendar_change(da_3d, caplog):
    """Ensure calendar logs change from 360 to 360_day."""
    caplog.set_level(logging.INFO)
    da_3d[da_3d.aat.t_dim].attrs["calendar"] = "360"
    da_3d.aat.correct_calendar()
    assert "Calendar attribute changed from '360' to '360_day'." in caplog.text


def test_correct_calendar_add(da_3d, caplog):
    """Ensure calendar logs change from units to 360_day."""
    caplog.set_level(logging.INFO)
    da_3d[da_3d.aat.t_dim].attrs["units"] = "months since 1960-01-01"
    da_3d.aat.correct_calendar()
    assert (
        "Calendar attribute '360_day' added, "
        "equivalent of 'units' 'months since'." in caplog.text
    )


def test_correct_calendar_no_change(da_3d, caplog):
    """Ensure calendar logs no change."""
    caplog.set_level(logging.INFO)
    da_3d[da_3d.aat.t_dim].attrs["calendar"] = "360_day"
    da_3d.aat.correct_calendar()
    assert "No 'units' or 'calendar' attributes to correct." in caplog.text


def test_change_longitude_range_no_change(da_2d, caplog):
    """Ensure change longitude logs no change."""
    caplog.set_level(logging.INFO)
    da_2d.aat.change_longitude_range()
    assert "Coordinates already in required range." in caplog.text


def test_compute_raster_stats_no_data_in_bounds(
    da_2d, gdf_missing, expected_2d, caplog
):
    """Ensure compute stats skips non-overlapping areas correctly."""
    caplog.set_level(logging.INFO)
    result = da_2d.aat.compute_raster_stats(
        gdf=gdf_missing, feature_col="name"
    )
    assert "No overlapping raster cells for area_c, skipping." in caplog.text
    assert_frame_equal(result, expected_2d, check_dtype=False)


def test_get_raster_array_crs(ds_3d):
    """Ensure dataset to array works when CRS set."""
    da = ds_3d.aat.get_raster_array("val")
    assert da.rio.crs == ds_3d.rio.crs



"""Test COD AB methods."""
import pytest

from aatoolbox import CodAB

DATASOURCE_BASE_DIR = "cod_ab"


@pytest.fixture
def downloader(mocker):
    """Mock the HDX download function."""
    return mocker.patch(
        "aatoolbox.datasources.codab.codab.load_dataset_from_hdx"
    )


@pytest.fixture
def gpd_read_file(mocker):
    """Mock GeoPandas file reading function."""
    return mocker.patch("aatoolbox.datasources.codab.codab.gpd.read_file")


def test_codab_download(mock_aa_data_dir, mock_country_config, downloader):
    """Test that load_codab calls the HDX API to download."""
    codab = CodAB(country_config=mock_country_config)
    codab.download()
    downloader.assert_called_with(
        hdx_address=f"cod-ab-{mock_country_config.iso3}",
        hdx_dataset_name=mock_country_config.codab.hdx_dataset_name,
        output_filepath=mock_aa_data_dir
        / f"public/raw/{mock_country_config.iso3}/"
        f"{DATASOURCE_BASE_DIR}/{mock_country_config.iso3}_"
        f"{DATASOURCE_BASE_DIR}.shp.zip",
    )


def test_codab_load_admin_level(
    mock_aa_data_dir, mock_country_config, gpd_read_file
):
    """Test that load_codab retrieves expected file and layer name."""
    codab = CodAB(country_config=mock_country_config)

    # First checking layer_base_name
    expected_layer_name = "fake_layer_base_name_level1"
    codab.load(admin_level=1)

    gpd_read_file.assert_called_with(
        f"zip://{mock_aa_data_dir}/public/raw/{mock_country_config.iso3}/"
        f"{DATASOURCE_BASE_DIR}/{mock_country_config.iso3}_"
        f"{DATASOURCE_BASE_DIR}.shp.zip/{expected_layer_name}"
    )

    # Then checking custom name
    expected_layer_name = "admin2_custom_name"
    codab.load(admin_level=2)

    gpd_read_file.assert_called_with(
        f"zip://{mock_aa_data_dir}/public/raw/{mock_country_config.iso3}/"
        f"{DATASOURCE_BASE_DIR}/{mock_country_config.iso3}_"
        f"{DATASOURCE_BASE_DIR}.shp.zip/{expected_layer_name}"
    )


def test_codab_too_high_admin_level(mock_country_config):
    """Test raised error when too high admin level requested."""
    codab = CodAB(country_config=mock_country_config)
    with pytest.raises(AttributeError):
        codab.load(admin_level=10)


def test_codab_custom(mock_aa_data_dir, mock_country_config, gpd_read_file):
    """Test that load_codab_custom retrieves expected file and layer name."""
    custom_layer_number = 1
    custom_layer_name_list = ["custom_layer_A", "custom_layer_B"]
    mock_country_config.codab.custom_layer_names = custom_layer_name_list
    codab = CodAB(country_config=mock_country_config)
    codab.load_custom(custom_layer_number)
    gpd_read_file.assert_called_with(
        f"zip://{mock_aa_data_dir}/public/raw/{mock_country_config.iso3}/"
        f"{DATASOURCE_BASE_DIR}/{mock_country_config.iso3}_"
        f"{DATASOURCE_BASE_DIR}.shp.zip/"
        f"{custom_layer_name_list[custom_layer_number]}"
    )


def test_codab_custom_missing(mock_country_config, gpd_read_file):
    """Test raised error when custom COD AB missing."""
    codab = CodAB(country_config=mock_country_config)
    with pytest.raises(AttributeError):
        codab.load_custom(0)


def test_codab_load_fail(mock_country_config):
    """Test raises file not found error when load fails."""
    codab = CodAB(country_config=mock_country_config)
    # Remove file if it exists
    # TODO: Use missing_ok=True once 3.7 is dropped
    if codab._raw_filepath.exists():
        codab._raw_filepath.unlink()
    with pytest.raises(FileNotFoundError) as excinfo:
        codab.load(admin_level=0)
    assert (
        "Make sure that you have already called the 'download' method"
        in str(excinfo.value)
    )
